"""Base class for chefs."""

import asyncio
import os
from abc import ABC, abstractmethod

from chonkie.logger import get_logger
from chonkie.types import Document

logger = get_logger(__name__)


class BaseChef(ABC):
    """Base class for chefs."""

    @abstractmethod
    def process(self, path: str | os.PathLike) -> Document:
        """Process the data from a file path.

        Args:
            path: Path to the file to process.

        Returns:
            Document created from the file.

        """
        raise NotImplementedError("Subclasses must implement process()")

    async def aprocess(self, path: str | os.PathLike) -> Document:
        """Process the data from a file path asynchronously."""
        return await asyncio.to_thread(self.process, path)

    @abstractmethod
    def parse(self, text: str) -> Document:
        """Parse raw text into a Document.

        Args:
            text: Raw text to parse.

        Returns:
            Document created from the text.

        """
        raise NotImplementedError("Subclasses must implement parse()")

    async def aparse(self, text: str) -> Document:
        """Parse raw text into a Document asynchronously."""
        return await asyncio.to_thread(self.parse, text)

    def process_batch(self, paths: list[str | os.PathLike]) -> list[Document]:
        """Process multiple files in a batch.

        Args:
            paths: List of file paths to process.

        Returns:
            List of Documents created from the files.

        """
        logger.info(f"Processing batch of {len(paths)} files")
        results = [self.process(path) for path in paths]
        logger.info(f"Completed batch processing of {len(paths)} files")
        return results

    async def aprocess_batch(self, paths: list[str | os.PathLike]) -> list[Document]:
        """Process multiple files in a batch asynchronously."""
        return await asyncio.gather(*[self.aprocess(path) for path in paths])

    def read(self, path: str | os.PathLike) -> str:
        """Read the file content.

        Args:
            path: Path to the file to read.

        Returns:
            File content as string.

        """
        try:
            logger.debug(f"Reading file: {path}")
            with open(path, "r", encoding="utf-8") as file:
                content = str(file.read())
                logger.debug(f"Successfully read file: {path}", size=len(content))
                return content
        except Exception as e:
            logger.warning(f"Failed to read file {path}: {e}", exc_info=True)
            raise

    async def aread(self, path: str | os.PathLike) -> str:
        """Read the file content asynchronously."""
        return await asyncio.to_thread(self.read, path)

    def __call__(self, path: str | os.PathLike) -> Document:
        """Call the chef to process the data.

        Args:
            path: Path to the file to process.

        Returns:
            Document created from the file.

        """
        logger.debug(f"Processing file with {self.__class__.__name__}: {path}")
        return self.process(path)

    def __repr__(self) -> str:
        """Return a string representation of the chef."""
        return f"{self.__class__.__name__}()"



"""Module for Chonkie's Porters.

Porters allow the user to _export_ data from chonkie into a variety of formats for saving on disk or cloud blob storage. Porters make the implicit assumption that the data is not being used for querying, but rather for saving.
"""

import asyncio
from abc import ABC, abstractmethod
from typing import Any

from chonkie.types import Chunk


class BasePorter(ABC):
    """Abstract base class for Chonkie's Porters.

    Porters are responsible for exporting Chonkie's Chunks into a variety of formats.
    The main method to implement is `export` or `aexport`, which should take in a list of Chunks
    and any other arguments, and export them to the desired format.
    """

    @abstractmethod
    def export(self, chunks: list[Chunk], **kwargs: dict[str, Any]) -> None:
        """Export the chunks to the desired format.

        Args:
            chunks: The chunks to export.
            **kwargs: Additional keyword arguments.

        Raises:
            NotImplementedError: If the subclass does not implement this method.

        """
        raise NotImplementedError("Subclasses must implement this method.")

    async def aexport(self, chunks: list[Chunk], **kwargs: dict[str, Any]) -> None:
        """Export the chunks to the desired format asynchronously.

        Args:
            chunks: The chunks to export.
            **kwargs: Additional keyword arguments.

        """
        return await asyncio.to_thread(self.export, chunks, **kwargs)

    def __call__(self, chunks: list[Chunk], **kwargs: dict[str, Any]) -> None:
        """Export the chunks to the desired format."""
        return self.export(chunks)



"""Module containing NeuralChunker class.

This module provides a NeuralChunker class for splitting text into chunks using a 100% neural approach, inspired by the work of [Chonky](https://github.com/mirth/chonky).

It trains an encoder style model on the task of token-classification (think: NER) to predict the split points of a text.
"""

import importlib.util as importutil
from typing import Any, Optional, Union, cast

from chonkie.logger import get_logger
from chonkie.pipeline import chunker
from chonkie.tokenizer import TokenizerProtocol
from chonkie.types import Chunk

from .base import BaseChunker

logger = get_logger(__name__)


@chunker("neural")
class NeuralChunker(BaseChunker):
    """Class for chunking text using a complete Neural Approach.

    This has been adapted from the implementation and models provided
    by [Chonky](https://github.com/mirth/chonky). This approach uses
    a token classification model to predict the split points in a text.

    Args:
      model: The model to use for the chunker.
      device: The device to use for the chunker.
      min_characters_per_chunk: The minimum number of characters per chunk.

    """

    SUPPORTED_MODELS = [
        "mirth/chonky_distilbert_base_uncased_1",
        "mirth/chonky_modernbert_base_1",
        "mirth/chonky_modernbert_large_1",
    ]

    SUPPORTED_MODEL_STRIDES = {
        "mirth/chonky_distilbert_base_uncased_1": 256,
        "mirth/chonky_modernbert_base_1": 512,
        "mirth/chonky_modernbert_large_1": 512,
    }

    DEFAULT_MODEL = "mirth/chonky_distilbert_base_uncased_1"

    def __init__(
        self,
        model: Union[str, Any] = DEFAULT_MODEL,
        tokenizer: Optional[Union[str, Any]] = None,
        device_map: str = "auto",
        min_characters_per_chunk: int = 10,
        stride: Optional[int] = None,
    ) -> None:
        """Initialize the NeuralChunker object.

        Args:
          model: The model to use for the chunker.
          tokenizer: The tokenizer to use for the chunker.
          device_map: The device to use for the chunker.
          min_characters_per_chunk: The minimum number of characters per chunk.
          stride: The stride to use for the chunker.

        """
        try:
            from transformers import (
                AutoModelForTokenClassification,
                AutoTokenizer,
                PreTrainedTokenizerFast,
                pipeline,
            )
        except ImportError as e:
            raise ImportError(
                "transformers is not installed. Please install it with `pip install chonkie[neural]`.",
            ) from e
        # Initialize the tokenizer to pass in to the parent class
        try:
            if isinstance(tokenizer, str):
                tokenizer = AutoTokenizer.from_pretrained(tokenizer)
            elif tokenizer is None and isinstance(model, str):
                tokenizer = AutoTokenizer.from_pretrained(model)
            elif isinstance(tokenizer, PreTrainedTokenizerFast):
                tokenizer = tokenizer
            else:
                raise ValueError(
                    "Invalid tokenizer provided. Please provide a string or a transformers.PreTrainedTokenizerFast object.",
                )
        except Exception as e:
            raise ValueError(f"Error initializing tokenizer: {e}") from e

        # Initialize the Parent class with the tokenizer
        super().__init__(cast(TokenizerProtocol, tokenizer))

        # Initialize the model and stride
        try:
            if isinstance(model, str):
                # Check if the model is supported
                if model not in self.SUPPORTED_MODELS:
                    raise ValueError(
                        f"Model {model} is not supported. Please choose from one of the following: {self.SUPPORTED_MODELS}",
                    )
                # Initialize the model
                self.model = AutoModelForTokenClassification.from_pretrained(
                    model, device_map=device_map
                )
                # Set the stride
                stride = self.SUPPORTED_MODEL_STRIDES[model] if stride is None else stride
            elif model is not None and "transformers" in str(type(model)):
                # Assuming that the model is a transformers model already, since it has transformers in the name, teehee~
                self.model = model
                # Since a custom model is provided, we need to set the stride to 0
                stride = 0 if stride is None else stride
            else:
                raise ValueError(
                    "Invalid model provided. Please provide a string or a transformers.AutoModelForTokenClassification object.",
                )
        except Exception as e:
            raise ValueError(f"Error initializing model: {e}") from e

        # Set the attributes
        self.min_characters_per_chunk = min_characters_per_chunk

        # Initialize the pipeline
        try:
            self.pipe = pipeline(
                "token-classification",
                model=model,
                tokenizer=tokenizer,
                device_map=device_map,
                aggregation_strategy="simple",
                stride=stride,
            )
        except Exception as e:
            raise ValueError(f"Error initializing pipeline: {e}") from e

        # Set the _use_multiprocessing value to be False
        self._use_multiprocessing = False

    @classmethod
    def _is_available(cls) -> bool:
        """Check if the dependencies are installed."""
        return importutil.find_spec("transformers") is not None

    def _get_splits(self, response: list[dict[str, Any]], text: str) -> list[str]:
        """Get the text splits from the model."""
        splits = []
        current_index = 0
        for sample in response:
            splits.append(text[current_index : sample["end"]])
            current_index = sample["end"]
        if current_index < len(text):
            splits.append(text[current_index:])
        return splits

    def _merge_close_spans(self, response: list[dict[str, Any]]) -> list[dict[str, Any]]:
        """Replace the split points that are too close to each other with the last span."""
        if not response:
            return []

        merged_response = [response[0]]
        for i in range(1, len(response)):
            current_span = response[i]
            last_merged_span = merged_response[-1]

            if current_span["start"] - last_merged_span["end"] < self.min_characters_per_chunk:
                # If the current span is too close to the last merged one,
                # replace the last one with the current one.
                merged_response[-1] = current_span
            else:
                # Otherwise, append the current span.
                merged_response.append(current_span)

        return merged_response

    def _get_chunks_from_splits(self, splits: list[str]) -> list[Chunk]:
        """Create a list of Chunks from the splits."""
        chunks = []
        current_index = 0
        token_counts = self.tokenizer.count_tokens_batch(splits)
        for split, token_count in zip(splits, token_counts):
            chunks.append(
                Chunk(
                    text=split,
                    start_index=current_index,
                    end_index=current_index + len(split),
                    token_count=token_count,
                ),
            )
            current_index += len(split)
        return chunks

    def chunk(self, text: str) -> list[Chunk]:
        """Chunk the text into a list of chunks.

        Args:
          text: The text to chunk.

        Returns:
          A list of chunks.

        """
        logger.debug(f"Starting neural chunking for text of length {len(text)}")
        # Get the spans
        spans = self.pipe(text)
        logger.debug(f"Model predicted {len(spans)} split points")

        # Merge close spans, since the model sometimes predicts spans that are too close to each other
        # and we want to ensure that we don't have chunks that are too small
        merged_spans = self._merge_close_spans(spans)

        # Get the splits from the merged spans
        splits = self._get_splits(merged_spans, text)

        # Return the chunks
        chunks = self._get_chunks_from_splits(splits)
        logger.info(f"Created {len(chunks)} chunks using neural token classification")
        return chunks

    def __repr__(self) -> str:
        """Return the string representation of the object."""
        return (
            f"NeuralChunker(model={self.model},"
            f"tokenizer={self.tokenizer}, "
            f"min_characters_per_chunk={self.min_characters_per_chunk})"
        )



"""Weaviate Handshake to export Chonkie's Chunks into a Weaviate collection."""

import importlib.util as importutil
from typing import (
    TYPE_CHECKING,
    Any,
    Literal,
    Optional,
    Union,
)
from urllib.parse import urlparse
from uuid import NAMESPACE_OID, uuid5

from flatbuffers.builder import np

from chonkie.embeddings import AutoEmbeddings, BaseEmbeddings
from chonkie.logger import get_logger
from chonkie.pipeline import handshake
from chonkie.types import Chunk

from .base import BaseHandshake
from .utils import generate_random_collection_name

logger = get_logger(__name__)

if TYPE_CHECKING:
    from weaviate import WeaviateClient


@handshake("weaviate")
class WeaviateHandshake(BaseHandshake):
    """Weaviate Handshake to export Chonkie's Chunks into a Weaviate collection.

    This handshake allows storing Chonkie chunks in Weaviate with vector embeddings.
    It supports both API key and OAuth authentication methods.

    Args:
        client: Optional[weaviate.WeaviateClient]: An existing Weaviate client instance.
        collection_name: Union[str, Literal["random"]]: The name of the collection to use.
        embedding_model: Union[str, BaseEmbeddings]: The embedding model to use.
        url: Optional[str]: The URL to the Weaviate server.
        api_key: Optional[str]: The API key for authentication.
        auth_config: Optional[dict[str, Any]]: OAuth configuration for authentication.
        batch_size: int: The batch size for batch operations. Defaults to 100.
        batch_dynamic: bool: Whether to use dynamic batching. Defaults to True.
        batch_timeout_retries: int: Number of retries for batch timeouts. Defaults to 3.
        additional_headers: Optional[dict[str, str]]: Additional headers for the Weaviate client.

    """

    def __init__(
        self,
        client: Optional["WeaviateClient"] = None,
        collection_name: Union[str, Literal["random"]] = "random",
        embedding_model: Union[str, BaseEmbeddings] = "minishlab/potion-retrieval-32M",
        url: Optional[str] = None,
        api_key: Optional[str] = None,
        auth_config: Optional[dict[str, Any]] = None,
        batch_size: int = 100,
        batch_dynamic: bool = True,
        batch_timeout_retries: int = 3,
        additional_headers: Optional[dict[str, str]] = None,
        http_secure: bool = False,
        grpc_host: Optional[str] = None,
        grpc_port: int = 50051,
        grpc_secure: bool = False,
    ) -> None:
        """Initialize the Weaviate Handshake.

        Args:
            client: Optional[weaviate.Client]: An existing Weaviate client instance.
            collection_name: Union[str, Literal["random"]]: The name of the collection to use.
            embedding_model: Union[str, BaseEmbeddings]: The embedding model to use.
            url: Optional[str]: The URL to the Weaviate server.
            api_key: Optional[str]: The API key for authentication.
            auth_config: Optional[dict[str, Any]]: OAuth configuration for authentication.
            batch_size: int: The batch size for batch operations. Defaults to 100.
            batch_dynamic: bool: Whether to use dynamic batching. Defaults to True.
            batch_timeout_retries: int: Number of retries for batch timeouts. Defaults to 3.
            additional_headers: Optional[dict[str, str]]: Additional headers for the Weaviate client.
            http_secure: bool: Whether to use HTTPS for HTTP connections. Defaults to False.
            grpc_host: Optional[str]: The host for gRPC connections. Defaults to the same as HTTP host.
            grpc_port: int: The port for gRPC connections. Defaults to 50051.
            grpc_secure: bool: Whether to use a secure channel for gRPC connections. Defaults to False.

        """
        super().__init__()

        try:
            import weaviate
        except ImportError as ie:
            raise ImportError(
                "Weaviate not available. Please install it with `pip install chonkie[weaviate]`."
            ) from ie

        # Initialize the Weaviate client
        if client is None:
            if url is None:
                url = "http://localhost:8080"

            try:
                # connect to cloud client
                self.client = weaviate.connect_to_weaviate_cloud(
                    cluster_url=url,
                    auth_credentials=weaviate.auth.Auth.api_key(
                        api_key if api_key is not None else "",
                    ),
                )
            except Exception:
                # connect to a localhost
                # Parse the URL to get the host and port
                parsed_url = urlparse(url)
                host = parsed_url.hostname or "localhost"
                port = parsed_url.port or 8080

                auth_credentials: Optional[Any] = None
                if api_key is not None:
                    auth_credentials = weaviate.auth.Auth.api_key(api_key=api_key)
                elif auth_config is not None:
                    assert "client_secret" in auth_config, (
                        "client_secret is required in auth_config"
                    )
                    auth_credentials = weaviate.auth.Auth.client_credentials(
                        client_secret=auth_config.pop("client_secret"),
                        **auth_config,
                    )

                # Use provided grpc_host or default to HTTP host
                actual_grpc_host = grpc_host if grpc_host is not None else host

                self.client = weaviate.connect_to_custom(
                    http_host=host,
                    http_port=port,
                    http_secure=http_secure,
                    grpc_host=actual_grpc_host,
                    grpc_port=grpc_port,
                    grpc_secure=grpc_secure,
                    auth_credentials=auth_credentials,
                    headers=additional_headers,
                )
        else:
            self.client = client

        # Initialize the embedding model
        if isinstance(embedding_model, str):
            self.embedding_model = AutoEmbeddings.get_embeddings(embedding_model)
        elif isinstance(embedding_model, BaseEmbeddings):
            self.embedding_model = embedding_model
        else:
            raise ValueError("embedding_model must be a string or a BaseEmbeddings instance.")

        # Determine vector dimensions
        if (
            hasattr(self.embedding_model, "dimension")
            and self.embedding_model.dimension is not None
        ):
            self.vector_dimensions = self.embedding_model.dimension
        else:
            # Fall back to test embedding if dimension property is not available or is None
            test_embedding = self.embedding_model.embed("test")
            self.vector_dimensions = len(test_embedding)

        # Set up batch configuration
        self.batch_size = batch_size
        self.batch_dynamic = batch_dynamic
        self.batch_timeout_retries = batch_timeout_retries

        # Initialize the collection
        if collection_name == "random":
            while True:
                self.collection_name = generate_random_collection_name(sep="_")
                # Check if the collection exists
                if not self._collection_exists(self.collection_name):
                    break
            logger.info(f"Chonkie created a new collection in Weaviate: {self.collection_name}")
        else:
            self.collection_name = collection_name

        # Create the collection if it doesn't exist
        if not self._collection_exists(self.collection_name):
            self._create_collection()

    @classmethod
    def _is_available(cls) -> bool:
        """Check if the dependencies are available."""
        return importutil.find_spec("weaviate") is not None

    def close(self) -> None:
        """Close."""
        self.client.close()

    def _collection_exists(self, collection_name: str) -> bool:
        """Check if a collection exists in Weaviate.

        Args:
            collection_name: The name of the collection to check.

        Returns:
            bool: True if the collection exists, False otherwise.

        """
        try:
            exists = self.client.collections.exists(collection_name)
            return exists
        except Exception as e:
            logger.warning(
                f"Failed to check for collection '{collection_name}': {e}",
                exc_info=True,
            )
            return False

    def _create_collection(self) -> None:
        """Create a new collection in Weaviate."""
        from weaviate.collections.classes.config import Configure, DataType, Property

        try:
            # Define the collection schema
            self.client.collections.create(
                name=self.collection_name,
                vector_index_config=Configure.VectorIndex.hnsw(),
                properties=[
                    Property.model_validate({
                        "name": "text",
                        "data_type": DataType.TEXT,
                        "description": "The text content of the chunk",
                    }),
                    Property.model_validate({
                        "name": "start_index",
                        "data_type": DataType.INT,
                        "description": "The start index of the chunk in the original text",
                    }),
                    Property.model_validate({
                        "name": "end_index",
                        "data_type": DataType.INT,
                        "description": "The end index of the chunk in the original text",
                    }),
                    Property.model_validate({
                        "name": "token_count",
                        "data_type": DataType.INT,
                        "description": "The number of tokens in the chunk",
                    }),
                    Property.model_validate({
                        "name": "chunk_type",
                        "data_type": DataType.TEXT,
                        "description": "The type of the chunk",
                    }),
                ],
            )

            logger.info(f"Created Weaviate collection: {self.collection_name}")
        except Exception:
            raise

    def _generate_id(self, index: int, chunk: Chunk) -> str:
        """Generate a unique ID for the chunk.

        Args:
            index: The index of the chunk in the batch.
            chunk: The chunk to generate an ID for.

        Returns:
            str: A unique ID for the chunk.

        """
        return str(uuid5(NAMESPACE_OID, f"{self.collection_name}::chunk-{index}:{chunk.text}"))

    def _generate_properties(self, chunk: Chunk) -> dict[str, Any]:
        """Generate properties for the chunk.

        Args:
            chunk: The chunk to generate properties for.

        Returns:
            dict[str, Any]: The properties for the chunk.

        """
        properties = {
            "text": chunk.text,
            "start_index": chunk.start_index,
            "end_index": chunk.end_index,
            "token_count": chunk.token_count,
            "chunk_type": type(chunk).__name__,
        }

        # Add chunk-specific properties
        sentences = getattr(chunk, "sentences", None)
        if sentences:
            properties["sentence_count"] = len(sentences)

        words = getattr(chunk, "words", None)
        if words:
            properties["word_count"] = len(words)

        if hasattr(chunk, "language") and chunk.language:
            properties["language"] = chunk.language

        return properties

    def write(self, chunks: Union[Chunk, list[Chunk]]) -> list[str]:
        """Write chunks to the Weaviate collection.

        Args:
            chunks: A single chunk or sequence of chunks to write.

        Returns:
            list[str]: List of IDs of the inserted chunks.

        Raises:
            RuntimeError: If there are too many errors during batch processing.

        """
        if isinstance(chunks, Chunk):
            chunks = [chunks]
        elif not isinstance(chunks, list):
            chunks = list(chunks)

        logger.debug(
            f"Writing {len(chunks)} chunks to Weaviate collection: {self.collection_name}",
        )
        # Get the collection
        collection = self.client.collections.get(self.collection_name)

        # Create a batch
        with collection.batch.fixed_size(batch_size=self.batch_size) as batch:
            chunk_ids = []
            max_errors = min(len(chunks) // 10 + 1, 10)  # Allow up to 10% errors or max 10

            for index, chunk in enumerate(chunks):
                # Check if we've hit too many errors
                if batch.number_errors > max_errors:
                    error_msg = f"Too many errors during batch processing ({batch.number_errors}). Aborting."
                    logger.error(error_msg)

                    raise RuntimeError(error_msg)

                try:
                    # Generate ID and properties
                    chunk_id = self._generate_id(index, chunk)
                    properties = self._generate_properties(chunk)

                    # Generate embedding
                    embedding = self.embedding_model.embed(chunk.text)

                    vector: list[float]
                    if isinstance(embedding, np.ndarray):
                        vector = embedding.tolist()
                    else:
                        vector = list(embedding)

                    # Add to batch
                    batch.add_object(properties=properties, uuid=chunk_id, vector=vector)

                    chunk_ids.append(chunk_id)
                except Exception as e:
                    logger.error(f"Error processing chunk {index}: {e}", exc_info=True)
                    # Continue with next chunk

            # After batch is complete, check for errors
            if batch.number_errors > 0:
                logger.warning(f"Completed with {batch.number_errors} errors")

        failed_objects = collection.batch.failed_objects
        if failed_objects:
            logger.warning(f"Number of failed imports: {len(failed_objects)}")
            if len(failed_objects) > 0:
                logger.error(f"First failed object: {failed_objects[0]}")

        # Report success
        successful_chunks = len(chunk_ids)
        logger.info(
            f"Chonkie wrote {successful_chunks} chunks to Weaviate collection: {self.collection_name}",
        )
        if successful_chunks < len(chunks):
            logger.warning(f"{len(chunks) - successful_chunks} chunks failed to write")

        return chunk_ids

    def delete_collection(self) -> None:
        """Delete the entire collection."""
        if self._collection_exists(self.collection_name):
            self.client.collections.delete(self.collection_name)
            logger.info(f"Deleted Weaviate collection: {self.collection_name}")

    def get_collection_info(self) -> dict[str, Any]:
        """Get information about the collection.

        Returns:
            dict[str, Any]: Information about the collection.

        """
        if not self._collection_exists(self.collection_name):
            return {"name": self.collection_name, "exists": False}

        collection = self.client.collections.get(self.collection_name)
        schema = collection.config.get()

        # Get property names list
        property_names: list[str] = []
        default_properties = [
            "text",
            "start_index",
            "end_index",
            "token_count",
            "chunk_type",
        ]

        if hasattr(schema, "properties") and schema.properties:
            try:
                # Handle both real properties and mock objects in tests
                property_names = []
                for prop in schema.properties:
                    # For test mocks, the name attribute might be a Mock itself
                    if hasattr(prop, "name"):
                        if isinstance(prop.name, str):
                            property_names.append(prop.name)
                        else:
                            # If it's a Mock or other non-string, use the attribute name
                            # This works for test mocks like Mock(name="text")
                            property_names.append(str(prop).split("name='")[1].split("'")[0])
            except (AttributeError, TypeError, IndexError):
                # Fallback to default properties if we can't get names
                property_names = default_properties

        # Use default property list if empty
        if not property_names:
            property_names = default_properties

        return {
            "name": self.collection_name,
            "exists": True,
            "vector_dimensions": self.vector_dimensions,
            "properties": property_names,
        }

    def __repr__(self) -> str:
        """Return the string representation of the WeaviateHandshake."""
        return f"WeaviateHandshake(collection_name={self.collection_name}, vector_dimensions={self.vector_dimensions})"

    def search(
        self,
        query: Optional[str] = None,
        embedding: Optional[list[float]] = None,
        limit: int = 5,
    ) -> list[dict[str, Any]]:
        """Retrieve the top_k most similar chunks to the query.

        Args:
            query: Optional[str]: The query string to search for.
            embedding: Optional[list[float]]: The embedding vector to search for. If provided, `query` is ignored.
            limit: int: The number of top similar chunks to retrieve.

        Returns:
            list[dict[str, Any]]: The list of most similar chunks with their metadata.

        """
        from weaviate.classes.query import MetadataQuery

        logger.debug(f"Searching Weaviate collection: {self.collection_name} with limit={limit}")
        if embedding is None and query is None:
            raise ValueError("Either query or embedding must be provided")
        if query is not None:
            embedding = self.embedding_model.embed(query).tolist()
        # enforce typing
        if not isinstance(embedding, list) or not all(isinstance(x, float) for x in embedding):
            raise ValueError("Embedding must be a list of floats")
        collection = self.client.collections.get(self.collection_name)
        # Weaviate expects a vector for similarity search
        results = collection.query.near_vector(
            near_vector=embedding,
            limit=limit,
            return_metadata=MetadataQuery(distance=True),
        )
        # Format results to match other handshakes
        matches = []
        for obj in results.objects:
            score = getattr(obj.metadata, "distance", None) if obj.metadata else None
            # Weaviate returns distance, convert to similarity (1 - distance) if needed
            similarity = 1.0 - score if score is not None else None
            match = {
                "id": obj.uuid,
                "score": similarity,
                "text": obj.properties.get("text"),
                "start_index": obj.properties.get("start_index"),
                "end_index": obj.properties.get("end_index"),
                "token_count": obj.properties.get("token_count"),
                "chunk_type": obj.properties.get("chunk_type"),
            }
            matches.append(match)
        logger.info(f"Search complete: found {len(matches)} matching chunks")
        return matches



"""Component for pipeline."""

from dataclasses import dataclass
from enum import Enum
from typing import Any


class ComponentType(Enum):
    """Types of pipeline components.

    These represent the stages in the CHOMP pipeline:
    - FETCHER: Retrieves raw data from sources
    - CHEF: Preprocesses and transforms data
    - CHUNKER: Splits text into chunks
    - REFINERY: Post-processes chunks (e.g., add embeddings, merge)
    - PORTER: Exports chunks to storage formats
    - HANDSHAKE: Ingests chunks into vector databases
    """

    FETCHER = "fetcher"
    CHEF = "chef"
    CHUNKER = "chunker"
    REFINERY = "refinery"
    PORTER = "porter"
    HANDSHAKE = "handshake"


@dataclass
class Component:
    """Minimal metadata about a pipeline component.

    This class stores the essential information needed to identify
    and instantiate components in a Chonkie pipeline.

    Attributes:
        name: Full class name (e.g., "RecursiveChunker")
        alias: Short alias for string-based configs (e.g., "recursive")
        component_class: The actual class to instantiate
        component_type: Which CHOMP stage this component belongs to

    """

    name: str
    alias: str
    component_class: type[Any]
    component_type: ComponentType

    def __post_init__(self) -> None:
        """Validate component after creation."""
        if not self.name:
            raise ValueError("Component name cannot be empty")
        if not self.alias:
            raise ValueError("Component alias cannot be empty")
        if not self.component_class:
            raise ValueError("Component class cannot be None")



"""Base Class for All Chunkers."""

import asyncio
from abc import ABC, abstractmethod
from dataclasses import replace
from typing import Sequence, Union

from tqdm import tqdm

from chonkie.logger import get_logger
from chonkie.tokenizer import AutoTokenizer, TokenizerProtocol
from chonkie.types import Chunk, Document

logger = get_logger(__name__)


class BaseChunker(ABC):
    """Base class for all chunkers."""

    def __init__(self, tokenizer: Union[str, TokenizerProtocol] = "gpt2"):
        """Initialize the chunker with any necessary parameters.

        Args:
            tokenizer: The tokenizer to use. Can be:
                - A string identifier (e.g., "gpt2", "character", "word")
                - An object implementing TokenizerProtocol (encode, decode, tokenize methods)

        """
        self._tokenizer = AutoTokenizer(tokenizer)
        self._use_multiprocessing = False
        logger.debug(
            f"Initialized {self.__class__.__name__}",
            tokenizer=str(tokenizer)[:50],
        )

    @property
    def tokenizer(self) -> AutoTokenizer:
        """Get the tokenizer instance."""
        return self._tokenizer

    def __repr__(self) -> str:
        """Return a string representation of the chunker."""
        return f"{self.__class__.__name__}()"

    def __call__(
        self,
        text: Union[str, Sequence[str]],
        show_progress: bool = True,
    ) -> Union[list[Chunk], list[list[Chunk]]]:
        """Call the chunker with the given arguments.

        Args:
            text (Union[str, Sequence[str]]): The text to chunk.
            show_progress (bool): Whether to show progress.

        Returns:
            If the input is a string, return a list of Chunks.
            If the input is a list of strings, return a list of lists of Chunks.

        """
        if isinstance(text, str):
            return self.chunk(text)
        elif isinstance(text, Sequence):
            return self.chunk_batch(text, show_progress)
        else:
            raise ValueError("Input must be a string or a list of strings.")

    def _get_optimal_worker_count(self) -> int:
        """Get the optimal number of workers for parallel processing."""
        try:
            from multiprocessing import cpu_count

            cpu_cores = cpu_count()
            worker_count = min(8, max(1, cpu_cores * 3 // 4))
            logger.debug(
                f"Using {worker_count} workers for parallel processing",
                cpu_cores=cpu_cores,
            )
            return worker_count
        except Exception:
            logger.warning(
                "Failed to calculate optimal worker count, using 1 worker",
                exc_info=True,
            )
            return 1

    def _sequential_batch_processing(
        self,
        texts: Sequence[str],
        show_progress: bool = True,
    ) -> list[list[Chunk]]:
        """Process a batch of texts sequentially."""
        logger.info(f"Starting sequential batch processing of {len(texts)} texts")
        results = [
            self.chunk(t)
            for t in tqdm(
                texts,
                desc="🦛",
                disable=not show_progress,
                unit="doc",
                bar_format="{desc} ch{bar:20}nk {percentage:3.0f}% • {n_fmt}/{total_fmt} docs chunked [{elapsed}<{remaining}, {rate_fmt}] 🌱",
                ascii=" o",
            )
        ]
        total_chunks = sum(len(r) for r in results)
        logger.info(
            f"Completed sequential processing: {total_chunks} total chunks from {len(texts)} texts",
        )
        return results

    def _parallel_batch_processing(
        self,
        texts: Sequence[str],
        show_progress: bool = True,
    ) -> list[list[Chunk]]:
        """Process a batch of texts using multiprocessing."""
        from multiprocessing import Pool

        num_workers = self._get_optimal_worker_count()
        total = len(texts)
        chunk_size = max(1, min(total // (num_workers * 16), 10))

        logger.info(
            f"Starting parallel batch processing of {total} texts",
            workers=num_workers,
            chunk_size=chunk_size,
        )

        with Pool(processes=num_workers) as pool:
            results = []
            with tqdm(
                total=total,
                desc="🦛",
                disable=not show_progress,
                unit="doc",
                bar_format="{desc} ch{bar:20}nk {percentage:3.0f}% • {n_fmt}/{total_fmt} docs chunked [{elapsed}<{remaining}, {rate_fmt}] 🌱",
                ascii=" o",
            ) as progress_bar:
                for result in pool.imap(self.chunk, texts, chunksize=chunk_size):
                    results.append(result)
                    progress_bar.update()

            total_chunks = sum(len(r) for r in results)
            logger.info(
                f"Completed parallel processing: {total_chunks} total chunks from {total} texts",
            )
            return results

    @abstractmethod
    def chunk(self, text: str) -> list[Chunk]:
        """Chunk the given text.

        Args:
            text (str): The text to chunk.

        Returns:
            list[Chunk]: A list of Chunks.

        """
        pass

    def chunk_batch(self, texts: Sequence[str], show_progress: bool = True) -> list[list[Chunk]]:
        """Chunk a batch of texts.

        Args:
            texts (Sequence[str]): The texts to chunk.
            show_progress (bool): Whether to show progress.

        Returns:
            list[list[Chunk]]: A list of lists of Chunks.

        """
        # simple handles of empty and single text cases
        if len(texts) == 0:
            return []
        if len(texts) == 1:
            return [self.chunk(texts[0])]

        # Now for the remaining, check the self._multiprocessing bool flag
        if self._use_multiprocessing:
            return self._parallel_batch_processing(texts, show_progress)
        else:
            return self._sequential_batch_processing(texts, show_progress)

    async def achunk(self, text: str) -> list[Chunk]:
        """Chunk the given text asynchronously.

        Args:
            text (str): The text to chunk.

        Returns:
            list[Chunk]: A list of Chunks.

        """
        return await asyncio.to_thread(self.chunk, text)

    async def achunk_batch(
        self, texts: Sequence[str], show_progress: bool = True
    ) -> list[list[Chunk]]:
        """Chunk a batch of texts asynchronously.

        Args:
            texts (Sequence[str]): The texts to chunk.
            show_progress (bool): Whether to show progress.

        Returns:
            list[list[Chunk]]: A list of lists of Chunks.

        """
        return await asyncio.to_thread(self.chunk_batch, texts, show_progress)

    @staticmethod
    def _merge_new_chunks(
        original_chunks: list[Chunk], new_chunk_batches: list[list[Chunk]]
    ) -> list[Chunk]:
        """Merge new chunks batches into a single list, shifting indices.

        Args:
            original_chunks: The original chunks from the document.
            new_chunk_batches: The new batches of chunks corresponding to each original chunk.

        Returns:
            list[Chunk]: The merged and shifted chunks.

        """
        return [
            replace(
                c,
                start_index=c.start_index + old_chunk.start_index,
                end_index=c.end_index + old_chunk.start_index,
            )
            for old_chunk, new_chunks in zip(original_chunks, new_chunk_batches)
            for c in new_chunks
        ]

    def chunk_document(self, document: Document) -> Document:
        """Chunk a document.

        Args:
            document: The document to chunk.

        Returns:
            The document with chunks populated.

        """
        # If the document has chunks already, then we need to re-chunk the content
        if document.chunks:
            chunk_results = [self.chunk(c.text) for c in document.chunks]
            document.chunks = self._merge_new_chunks(document.chunks, chunk_results)
        else:
            document.chunks = self.chunk(document.content)
        return document

    async def achunk_document(self, document: Document) -> Document:
        """Chunk a document asynchronously.

        Args:
            document: The document to chunk.

        Returns:
            The document with chunks populated.

        """
        # If the document has chunks already, then we need to re-chunk the content
        if document.chunks:
            tasks = [self.achunk(c.text) for c in document.chunks]
            chunk_results = await asyncio.gather(*tasks)
            document.chunks = self._merge_new_chunks(document.chunks, chunk_results)
        else:
            document.chunks = await self.achunk(document.content)
        return document



"""Utility helpers for the Chonkie OSS API.

Provides a simple Timer for performance logging and re-exports the Chonkie
logger for consistent logging across the entire package.
"""

import time

from chonkie.logger import get_logger

__all__ = ["get_logger", "Timer", "fix_escaped_text", "sanitize_text_encoding"]


class Timer:
    """Lightweight wall-clock timer with named sub-timers.

    Usage::

        timer = Timer()
        timer.start()           # start the global timer

        timer.start("init")     # start a named timer
        ...
        ms = timer.end("init")  # stop and return elapsed ms

        total_ms = timer.end()  # stop global timer
    """

    def __init__(self) -> None:
        """Initialize Timer with empty start times."""
        self._starts: dict[str, float] = {}

    def start(self, name: str = "_global") -> None:
        """Start (or restart) a named timer.

        Args:
            name: Timer name.  Defaults to the global timer.

        """
        self._starts[name] = time.perf_counter()

    def end(self, name: str = "_global") -> float:
        """Stop a named timer and return elapsed milliseconds.

        Args:
            name: Timer name.  Defaults to the global timer.

        Returns:
            Elapsed time in milliseconds, or ``0.0`` if the timer was never
            started.

        """
        start = self._starts.pop(name, None)
        if start is None:
            return 0.0
        return (time.perf_counter() - start) * 1000.0

    def elapsed(self, name: str = "_global") -> float:
        """Return elapsed milliseconds without stopping the timer.

        Args:
            name: Timer name.

        Returns:
            Elapsed time in milliseconds, or ``0.0`` if not started.

        """
        start = self._starts.get(name)
        if start is None:
            return 0.0
        return (time.perf_counter() - start) * 1000.0


def fix_escaped_text(text: str | list[str]) -> str | list[str]:
    """Unescape common escape sequences that arrive double-escaped over JSON.

    Args:
        text: A string or list of strings to fix.

    Returns:
        The fixed string or list of strings.

    """
    if isinstance(text, list):
        return [_fix_single(t) for t in text]
    return _fix_single(text)


def _fix_single(text: str) -> str:
    replacements = [
        ("\\n", "\n"),
        ("\\t", "\t"),
        ("\\r", "\r"),
    ]
    for escaped, real in replacements:
        if escaped in text:
            text = text.replace(escaped, real)
    return text


def sanitize_text_encoding(text: str) -> str:
    """Replace invalid UTF-8 byte sequences with the replacement character.

    Args:
        text: Input string.

    Returns:
        Sanitized string safe for processing.

    """
    return text.encode("utf-8", errors="replace").decode("utf-8", errors="replace")



"""VoyageAI embeddings - backward-compatible wrapper around CatsuEmbeddings.

Note: Consider using CatsuEmbeddings(model=..., provider="voyageai") directly.
"""

import importlib.util as importutil
import os
import warnings
from typing import Any, Literal, Optional

import numpy as np

from .base import BaseEmbeddings
from .catsu import CatsuEmbeddings


class VoyageAIEmbeddings(BaseEmbeddings):
    """VoyageAI embeddings via CatsuEmbeddings.

    This is a backward-compatible wrapper around CatsuEmbeddings.
    Consider using CatsuEmbeddings(model=..., provider="voyageai") directly.

    Args:
        model: VoyageAI embedding model name (default: "voyage-3").
        api_key: VoyageAI API key (falls back to VOYAGE_API_KEY or VOYAGEAI_API_KEY env var).
        max_retries: Maximum retry attempts (default: 3).
        timeout: Request timeout in seconds (default: 60).
        output_dimension: Ignored; kept for backward compatibility.
        batch_size: Number of texts per API call (default: 128).
        truncation: Ignored; kept for backward compatibility.

    """

    AVAILABLE_MODELS = {
        "voyage-3-large": ((1024, 256, 512, 2048), 32000),
        "voyage-3": ((1024,), 32000),
        "voyage-3-lite": ((512,), 32000),
        "voyage-code-3": ((1024, 256, 512, 2048), 32000),
        "voyage-finance-2": ((1024,), 32000),
        "voyage-law-2": ((1024,), 16000),
        "voyage-code-2": ((1536,), 16000),
    }

    DEFAULT_MODEL = "voyage-3"

    def __init__(
        self,
        model: str = DEFAULT_MODEL,
        api_key: Optional[str] = None,
        max_retries: int = 3,
        timeout: float = 60.0,
        output_dimension: Optional[Literal[256, 512, 1024, 2048]] = None,
        batch_size: int = 128,
        truncation: bool = True,
    ):
        """Initialize VoyageAI embeddings wrapper.

        Args:
            model: VoyageAI embedding model name.
            api_key: VoyageAI API key (falls back to VOYAGE_API_KEY or VOYAGEAI_API_KEY env var).
            max_retries: Maximum retry attempts.
            timeout: Request timeout in seconds.
            output_dimension: Ignored; kept for backward compatibility.
            batch_size: Number of texts per API call (max 128).
            truncation: Ignored; kept for backward compatibility.

        Raises:
            ImportError: If the catsu package is not installed.

        """
        super().__init__()

        if not self._is_available():
            raise ImportError(
                "One (or more) of the following packages is not available: catsu. "
                'Please install it via `pip install "chonkie[catsu]"`',
            )

        if output_dimension is not None:
            warnings.warn(
                "The `output_dimension` parameter is not supported in this version and will be ignored.",
                DeprecationWarning,
                stacklevel=2,
            )
        if truncation is not True:
            warnings.warn(
                "The `truncation` parameter is not supported in this version and will be ignored.",
                DeprecationWarning,
                stacklevel=2,
            )

        self.model = model
        api_key = api_key or os.getenv("VOYAGE_API_KEY") or os.getenv("VOYAGEAI_API_KEY")
        api_keys = {"voyageai": api_key} if api_key else None

        self._catsu = CatsuEmbeddings(
            model=model,
            provider="voyageai",
            api_keys=api_keys,
            max_retries=max_retries,
            timeout=round(timeout),
            batch_size=min(batch_size, 128),
        )

    def embed(self, text: str) -> np.ndarray:
        """Embed a single text string."""
        return self._catsu.embed(text)

    def embed_batch(self, texts: list) -> list:
        """Embed multiple texts using batched API calls."""
        return self._catsu.embed_batch(texts)

    async def aembed(self, text: str) -> np.ndarray:
        """Embed a single text string asynchronously."""
        return await self._catsu.aembed(text)

    async def aembed_batch(self, texts: list) -> list:
        """Embed multiple texts asynchronously using batched API calls."""
        return await self._catsu.aembed_batch(texts)

    @property
    def dimension(self) -> int:
        """Return the embedding dimension."""
        return self._catsu.dimension

    def get_tokenizer(self) -> Any:
        """Return a tokenizer object for token counting."""
        return self._catsu.get_tokenizer()

    @classmethod
    def _is_available(cls) -> bool:
        """Check if the catsu package is available."""
        return importutil.find_spec("catsu") is not None

    def __repr__(self) -> str:
        """Return a string representation of the VoyageAIEmbeddings object."""
        return f"VoyageAIEmbeddings(model={self.model})"



"""Azure OpenAI Genie."""

import importlib.util as importutil
from typing import TYPE_CHECKING, Any, Optional

from .base import BaseGenie

if TYPE_CHECKING:
    from pydantic import BaseModel


class AzureOpenAIGenie(BaseGenie):
    """Azure-hosted OpenAI Genie wrapper."""

    def __init__(
        self,
        azure_endpoint: str,
        model: str = "gpt-4o",
        deployment: Optional[str] = None,
        azure_api_key: Optional[str] = None,
        api_version: str = "2024-10-21",
    ):
        """Initialize AzureOpenAIGenie.

        Args:
            model (str): Logical model name, used for tokenization etc.
            deployment (Optional[str]): Azure deployment name (used for API call). Defaults to `model`.
            azure_endpoint (Optional[str]): Your Azure OpenAI endpoint.
            azure_api_key (Optional[str]): Your Azure OpenAI API key. If omitted, uses Entra ID.
            api_version (str): Required API version (default: "2024-10-21").

        """
        super().__init__()

        if not azure_endpoint:
            raise ValueError("`azure_endpoint` is required for Azure OpenAI.")

        self.model = model
        self._deployment = deployment or model
        self.api_version = api_version
        self.base_url = azure_endpoint

        try:
            from openai import AsyncAzureOpenAI, AzureOpenAI
        except ImportError as ie:
            raise ImportError(
                "openai is not available. Please install it via `pip install chonkie[azure-openai]`",
            ) from ie

        if azure_api_key:
            self.client = AzureOpenAI(
                azure_endpoint=azure_endpoint,
                api_key=azure_api_key,
                api_version=api_version,
            )
            self.async_client = AsyncAzureOpenAI(
                azure_endpoint=azure_endpoint,
                api_key=azure_api_key,
                api_version=api_version,
            )
        else:
            from azure.identity import DefaultAzureCredential, get_bearer_token_provider

            token_provider = get_bearer_token_provider(
                DefaultAzureCredential(),
                "https://cognitiveservices.azure.com/.default",
            )
            self.client = AzureOpenAI(
                azure_endpoint=azure_endpoint,
                azure_ad_token_provider=token_provider,
                api_version=api_version,
            )
            self.async_client = AsyncAzureOpenAI(
                azure_endpoint=azure_endpoint,
                azure_ad_token_provider=token_provider,
                api_version=api_version,
            )

    def generate(self, prompt: str) -> str:
        """Generate a plain-text response."""
        response = self.client.chat.completions.create(
            model=self._deployment,
            messages=[{"role": "user", "content": prompt}],
        )
        content = response.choices[0].message.content
        if content is None:
            raise ValueError("Azure OpenAI response content is None")
        return content

    async def agenerate(self, prompt: str) -> str:
        """Generate a response asynchronously."""
        response = await self.async_client.chat.completions.create(
            model=self._deployment,
            messages=[{"role": "user", "content": prompt}],
        )
        content = response.choices[0].message.content
        if content is None:
            raise ValueError("Azure OpenAI response content is None")
        return content

    def generate_json(self, prompt: str, schema: "BaseModel") -> dict[str, Any]:
        """Generate a structured JSON response based on a Pydantic schema."""
        response = self.client.beta.chat.completions.parse(
            model=self._deployment,
            messages=[{"role": "user", "content": prompt}],
            response_format=schema,  # type: ignore[arg-type]
        )
        content = response.choices[0].message.parsed
        if content is None:
            raise ValueError("Azure OpenAI response content is None")
        return content.model_dump()

    async def agenerate_json(self, prompt: str, schema: "BaseModel") -> dict[str, Any]:
        """Generate a structured JSON response asynchronously."""
        response = await self.async_client.beta.chat.completions.parse(
            model=self._deployment,
            messages=[{"role": "user", "content": prompt}],
            response_format=schema,  # type: ignore[arg-type]
        )
        content = response.choices[0].message.parsed
        if content is None:
            raise ValueError("Azure OpenAI response content is None")
        return content.model_dump()

    @classmethod
    def _is_available(cls) -> bool:
        return (
            importutil.find_spec("openai") is not None
            and importutil.find_spec("pydantic") is not None
            and importutil.find_spec("azure.identity") is not None
        )

    def __repr__(self) -> str:
        """Return a string representation of the AzureOpenAIGenie instance."""
        return f"AzureOpenAIGenie(model={self.model}, deployment={self._deployment}, endpoint={self.base_url})"



"""TableChef is a chef that processes tabular data from files (e.g., CSV, Excel) and markdown strings."""

import os
import re
from pathlib import Path
from typing import Union, cast

from chonkie.chef.base import BaseChef
from chonkie.logger import get_logger
from chonkie.pipeline import chef
from chonkie.types import Document, MarkdownDocument, MarkdownTable

logger = get_logger(__name__)


@chef("table")
class TableChef(BaseChef):
    """TableChef processes CSV files and returns pandas DataFrames."""

    def __init__(self) -> None:
        """Initialize TableChef with regex patterns for markdown and HTML tables."""
        self.table_pattern = re.compile(r"(\|.*?\n(?:\|[-: ]+\|.*?\n)?(?:\|.*?\n)+)")
        # Matches individual opening/closing <table> tags only — no DOTALL, no .*?
        # Depth tracking in _find_html_table_spans handles nesting and avoids ReDoS.
        self._html_table_tag_pattern = re.compile(r"<(/?)table\b[^>]*>", re.IGNORECASE)

    def parse(self, text: str) -> Document:
        """Parse raw markdown text and extract tables into a MarkdownDocument.

        Args:
            text: Raw markdown text.

        Returns:
            Document: MarkdownDocument with extracted tables.

        """
        logger.debug("Parsing markdown text for tables")
        tables = self.extract_tables_from_markdown(text)
        logger.info(f"Markdown table extraction complete: found {len(tables)} tables")
        return MarkdownDocument(content=text, tables=tables)

    def process(self, path: str | os.PathLike) -> Document:
        """Process a CSV/Excel file or markdown text into a MarkdownDocument.

        Args:
            path: Path to the CSV/Excel file, or markdown text string.

        Returns:
            Document: MarkdownDocument with extracted tables.

        """
        logger.debug(f"Processing table file/string: {path}")
        try:
            import pandas as pd
        except ImportError as e:
            raise ImportError(
                "Pandas is required to use TableChef. Please install it with `pip install chonkie[table]`.",
            ) from e
        # if file exists
        path_obj = Path(path)
        if path_obj.is_file():
            str_path = str(path)
            if str_path.endswith(".csv"):
                logger.debug("Processing CSV file")
                df = pd.read_csv(str_path)
                markdown = df.to_markdown(index=False) or ""
                logger.info(f"CSV processing complete: converted {len(df)} rows to markdown")
                # CSV always produces a single table
                table = MarkdownTable(content=markdown, start_index=0, end_index=len(markdown))
                return MarkdownDocument(content=markdown, tables=[table])
            elif str_path.endswith(".xls") or str_path.endswith(".xlsx"):
                logger.debug("Processing Excel file")
                all_df = pd.read_excel(str_path, sheet_name=None)
                tables: list[MarkdownTable] = []
                all_content = []
                for df in all_df.values():
                    text = df.to_markdown(index=False) or ""
                    all_content.append(text)
                    tables.append(MarkdownTable(content=text, start_index=0, end_index=len(text)))
                # Join all sheets with double newline
                content = "\n\n".join(all_content)
                logger.info(
                    f"Excel processing complete: converted {len(all_df)} sheets to markdown",
                )
                return MarkdownDocument(content=content, tables=tables)
        # Not a file, treat as markdown string and extract tables
        logger.debug("Extracting tables from markdown string")
        return self.parse(str(path))

    def process_batch(self, paths: list[str | os.PathLike]) -> list[Document]:
        """Process multiple CSV/Excel files or markdown texts.

        Args:
            paths: Paths to files or markdown text strings.

        Returns:
            list[Document]: List of MarkdownDocuments with extracted tables.

        """
        logger.debug(f"Processing batch of {len(paths)} files/strings")
        results = [self.process(path) for path in paths]
        logger.info(f"Completed batch processing of {len(paths)} files/strings")
        return results

    def __call__(  # type: ignore[override]
        self,
        path: str | os.PathLike | list[str | os.PathLike],
    ) -> Union[Document, list[Document]]:
        """Process a single file/text or a batch of files/texts.

        Args:
            path: Single file path, markdown text string, or list of paths/texts.

        Returns:
            Union[Document, list[Document]]: MarkdownDocument or list of MarkdownDocuments.

        """
        if isinstance(path, (list, tuple)):
            return self.process_batch(cast("list[str | os.PathLike]", list(path)))
        elif isinstance(path, (str, Path)):
            return self.process(path)
        else:
            raise TypeError(f"Unsupported type: {type(path)}")

    def _find_html_table_spans(self, text: str) -> list[tuple[int, int]]:
        """Return (start, end) character spans for each top-level HTML table.

        Uses depth tracking so nested <table> elements are included in their
        enclosing table's span rather than being returned as separate tables.
        The tag pattern [^>]* has a hard stop character and cannot backtrack
        catastrophically regardless of input length.
        """
        spans: list[tuple[int, int]] = []
        depth = 0
        start = -1
        for m in self._html_table_tag_pattern.finditer(text):
            if m.group(1):  # closing </table>
                if depth > 0:
                    depth -= 1
                    if depth == 0:
                        spans.append((start, m.end()))
            else:  # opening <table …>
                if depth == 0:
                    start = m.start()
                depth += 1
        return spans

    def extract_tables_from_markdown(self, markdown: str) -> list[MarkdownTable]:
        """Extract markdown and HTML tables from a markdown string.

        Args:
            markdown (str): The markdown text containing tables.

        Returns:
            list[MarkdownTable]: A list of MarkdownTable objects, each representing a table found in the input.

        """
        tables: list[MarkdownTable] = []

        # Find HTML tables first; they take priority when spans overlap
        html_tables: list[MarkdownTable] = []
        for start, end in self._find_html_table_spans(markdown):
            html_tables.append(
                MarkdownTable(
                    content=markdown[start:end],
                    start_index=start,
                    end_index=end,
                )
            )

        # Find markdown pipe tables, skipping any that overlap with an HTML match
        for match in self.table_pattern.finditer(markdown):
            start_index = match.start()
            end_index = match.end()
            if any(
                start_index < ht.end_index and ht.start_index < end_index for ht in html_tables
            ):
                continue
            tables.append(
                MarkdownTable(content=match.group(0), start_index=start_index, end_index=end_index)
            )

        tables.extend(html_tables)
        # Sort tables by their appearance in the text
        tables.sort(key=lambda x: x.start_index)
        return tables

    def __repr__(self) -> str:
        """Return a string representation of the chef."""
        return f"{self.__class__.__name__}()"



"""Base class for all refinery classes."""

from abc import ABC, abstractmethod
from typing import Any


class BaseRefinery(ABC):
    """Base class for all cloud refinery classes."""

    BASE_URL = "https://api.chonkie.ai"
    VERSION = "v1"

    @abstractmethod
    def refine(self, chunks: list[Any]) -> list[Any]:
        """Refine the chunks."""
        raise NotImplementedError("Subclasses must implement this method.")

    def __call__(self, chunks: list[Any]) -> list[Any]:
        """Call the refinery.

        Args:
            chunks: The chunks to refine.

        Returns:
            The refined chunks.

        """
        return self.refine(chunks)



"""Module containing the LateChunker class."""

import os
from typing import Any, Optional, Union, cast

import numpy as np

# Get all the Chonkie imports
from chonkie.chunker.recursive import RecursiveChunker
from chonkie.embeddings.sentence_transformer import SentenceTransformerEmbeddings
from chonkie.logger import get_logger
from chonkie.pipeline import chunker
from chonkie.tokenizer import TokenizerProtocol
from chonkie.types import Chunk, RecursiveRules

logger = get_logger(__name__)


@chunker("late")
class LateChunker(RecursiveChunker):
    """A chunker that chunks texts based on late interaction.

    This class extends the RecursiveChunker class and overrides its chunk method to implement late chunking.

    Args:
        embedding_model: The embedding model to use for chunking.
        chunk_size: The maximum size of each chunk.
        rules: Recursive rules to chunk by
        min_characters_per_chunk: Minimum number of characters in a single chunk

    """

    def __init__(
        self,
        embedding_model: Union[
            str,
            SentenceTransformerEmbeddings,
            Any,
        ] = "nomic-ai/modernbert-embed-base",
        chunk_size: int = 2048,
        rules: RecursiveRules = RecursiveRules(),
        min_characters_per_chunk: int = 24,
        **kwargs: Any,
    ) -> None:
        """Initialize the LateChunker.

        Args:
            embedding_model: The embedding model to use for chunking.
            chunk_size: The maximum size of each chunk.
            rules: The rules to use for chunking.
            min_characters_per_chunk: The minimum number of characters per chunk.
            **kwargs: Additional keyword arguments.

        """
        # set all the additional attributes
        if isinstance(embedding_model, SentenceTransformerEmbeddings):
            self.embedding_model = embedding_model
        elif isinstance(embedding_model, str):
            self.embedding_model = SentenceTransformerEmbeddings(model=embedding_model, **kwargs)
        else:
            raise ValueError(f"{embedding_model} is not a valid embedding model")

        # Probably the dependency hasn't been installed
        if self.embedding_model is None:
            raise ImportError(
                "Oh! seems like you're missing the proper dependency to run this chunker. Please install it using `pip install chonkie[st]`",
            )

        # Initialize the RecursiveChunker with the embedding_model's tokenizer
        super().__init__(
            tokenizer=cast(TokenizerProtocol, self.embedding_model.get_tokenizer()),
            chunk_size=chunk_size,
            rules=rules,
            min_characters_per_chunk=min_characters_per_chunk,
        )

        # Disable multiprocessing for this chunker
        self._use_multiprocessing = False

    @classmethod
    def from_recipe(  # type: ignore[override]
        cls,
        name: Optional[str] = "default",
        lang: Optional[str] = "en",
        path: str | os.PathLike | None = None,
        embedding_model: Union[
            str,
            SentenceTransformerEmbeddings,
        ] = "sentence-transformers/all-MiniLM-L6-v2",
        chunk_size: int = 2048,
        min_characters_per_chunk: int = 24,
        **kwargs: Any,
    ) -> "LateChunker":
        """Create a LateChunker from a recipe.

        Args:
            name: The name of the recipe to use.
            lang: The language that the recipe should support.
            path: The path to the recipe to use.
            embedding_model: The embedding model to use.
            chunk_size: The chunk size to use.
            min_characters_per_chunk: The minimum number of characters per chunk.
            **kwargs: Additional keyword arguments.

        Returns:
            LateChunker: The created LateChunker.

        Raises:
            ValueError: If the recipe is invalid or if the recipe is not found.

        """
        logger.info("Loading LateChunker recipe", recipe_name=name, lang=lang)
        # Create a hubbie instance
        rules = RecursiveRules.from_recipe(name, lang, path)
        logger.debug(f"Recipe loaded successfully with {len(rules.levels or [])} levels")
        return cls(
            embedding_model=embedding_model,
            chunk_size=chunk_size,
            rules=rules,
            min_characters_per_chunk=min_characters_per_chunk,
            **kwargs,
        )

    def _get_late_embeddings(
        self,
        token_embeddings: np.ndarray,
        token_counts: list[int],
    ) -> list[np.ndarray]:
        # Split the token embeddings into chunks based on the token counts
        embs = []
        cum_token_counts = np.cumsum([0] + token_counts)
        for i in range(len(token_counts)):
            embs.append(
                np.mean(
                    token_embeddings[cum_token_counts[i] : cum_token_counts[i + 1]],
                    axis=0,
                ),
            )
        return embs

    def chunk(self, text: str) -> list[Chunk]:
        """Chunk the text via LateChunking."""
        logger.debug(f"Starting late chunking for text of length {len(text)}")
        # This would first call upon the _recursive_chunk method
        # and then use the embedding model to get the token token_embeddings
        # Lastly, we would combine the methods together to create the LateChunk objects
        chunks = self._recursive_chunk(text)
        logger.debug(f"Created {len(chunks)} initial chunks from recursive splitting")
        token_embeddings = self.embedding_model.embed_as_tokens(text)

        # Get the token_counts for all the chunks
        token_counts = [c.token_count for c in chunks]

        # If fallback was used, token_embeddings may be fewer than sum(token_counts)
        if token_embeddings.shape[0] < sum(token_counts):
            # Fallback: use sentence embeddings for each chunk
            # Re-embed each chunk as a sentence embedding
            token_embeddings = np.array([self.embedding_model.embed(c.text) for c in chunks])
            token_counts = [1 for _ in chunks]

        # Validate the token_counts with the actual count
        if sum(token_counts) > token_embeddings.shape[0]:
            raise ValueError("The sum of token counts exceeds the number of tokens in the text")
        if sum(token_counts) < token_embeddings.shape[0]:
            diff = token_embeddings.shape[0] - sum(token_counts)
            token_counts[0] = token_counts[0] + diff // 2
            token_counts[-1] = token_counts[-1] + (diff - diff // 2)

        if sum(token_counts) != token_embeddings.shape[0]:
            raise ValueError(
                "The sum of token counts does not match the number of tokens in the text",
                f"Expected {token_embeddings.shape[0]}, got {sum(token_counts)}",
            )

        # Split the token embeddings into chunks based on the token counts
        late_embds = self._get_late_embeddings(token_embeddings, token_counts)

        # Wrap it all up in Chunks with embeddings
        result = []
        for chunk, token_count, embedding in zip(chunks, token_counts, late_embds):
            # Note: LateChunker always returns chunks, so chunk is always a Chunk
            result.append(
                Chunk(
                    text=chunk.text,
                    start_index=chunk.start_index,
                    end_index=chunk.end_index,
                    token_count=token_count,
                    embedding=embedding,
                ),
            )
        logger.info(f"Created {len(result)} chunks with late interaction embeddings")
        return result



"""AutoEmbeddings is a factory class for automatically loading embeddings."""

from typing import Any, Union, cast

from chonkie.logger import get_logger

from .base import BaseEmbeddings
from .registry import EmbeddingsRegistry

logger = get_logger(__name__)


class AutoEmbeddings:
    """Factory class for automatically loading embeddings.

    This class provides a factory interface for loading embeddings based on an
    identifier string. It will try to find a matching embeddings implementation
    based on the identifier and load it with the provided arguments.


    Examples:
        # Get sentence transformers embeddings
        embeddings = AutoEmbeddings.get_embeddings("sentence-transformers/all-MiniLM-L6-v2")

        # Get OpenAI embeddings
        embeddings = AutoEmbeddings.get_embeddings("openai://text-embedding-ada-002", api_key="...")

        # Get Anthropic embeddings
        embeddings = AutoEmbeddings.get_embeddings("anthropic://claude-v1", api_key="...")

         # Get Cohere embeddings
        embeddings = AutoEmbeddings.get_embeddings("cohere://embed-english-light-v3.0", api_key="...")

    """

    @classmethod
    def get_embeddings(
        cls,
        model: Union[str, BaseEmbeddings, Any],
        **kwargs: Any,
    ) -> BaseEmbeddings:
        """Get embeddings instance based on identifier.

        Args:
            model: Identifier for the embeddings (name, path, URL, etc.)
            **kwargs: Additional arguments passed to the embeddings constructor

        Returns:
            Initialized embeddings instance

        Raises:
            ValueError: If no suitable embeddings implementation is found

        Examples:
            # Get sentence transformers embeddings
            embeddings = AutoEmbeddings.get_embeddings("sentence-transformers/all-MiniLM-L6-v2")

            # Get OpenAI embeddings
            embeddings = AutoEmbeddings.get_embeddings("openai://text-embedding-ada-002", api_key="...")

            # Get Anthropic embeddings
            embeddings = AutoEmbeddings.get_embeddings("anthropic://claude-v1", api_key="...")

            # Get Cohere embeddings
            embeddings = AutoEmbeddings.get_embeddings("cohere://embed-english-light-v3.0", api_key="...")

        """
        # Load embeddings instance if already provided
        if isinstance(model, BaseEmbeddings):
            return model
        elif isinstance(model, str):
            # Initializing the embedding instance
            embeddings_instance = None

            # Check if the user passed in a provider alias
            if "://" in model:
                provider, model_name = model.split("://")
                embeddings_cls = EmbeddingsRegistry.get_provider(provider)
                if embeddings_cls:
                    try:
                        return cast(Any, embeddings_cls)(model_name, **kwargs)
                    except Exception as error:
                        raise ValueError(
                            f"Failed to load {model} with {embeddings_cls.__name__}, with error: {error}",
                        ) from error
                else:
                    raise ValueError(
                        f"No provider found for {provider}. Please check the provider name and try again.",
                    )
            else:
                # Try to find matching implementation via registry
                embeddings_cls = EmbeddingsRegistry.match(model)
                if embeddings_cls:
                    try:
                        # Try instantiating with the model identifier
                        embeddings_instance = cast(Any, embeddings_cls)(model, **kwargs)
                    except Exception as error:
                        logger.warning(
                            f"Failed to load {model} with {embeddings_cls.__name__}: {error}\n"
                            "Falling back to loading default provider model.",
                            exc_info=True,
                        )
                        try:
                            # Try instantiating with the default provider model without the model identifier
                            embeddings_instance = embeddings_cls(**kwargs)
                        except Exception as error:
                            logger.warning(
                                f"Failed to load the default model for {embeddings_cls.__name__}: {error}\n"
                                "Falling back to SentenceTransformerEmbeddings.",
                                exc_info=True,
                            )

            # If registry lookup and instantiation succeeded, return the instance
            if embeddings_instance:
                return embeddings_instance

            # If registry lookup and instantiation failed, return the default SentenceTransformerEmbeddings
            from .sentence_transformer import SentenceTransformerEmbeddings

            try:
                return SentenceTransformerEmbeddings(model, **kwargs)
            except Exception as e:
                raise ValueError(
                    f"Failed to load embeddings via SentenceTransformerEmbeddings after registry/fallback failure: {e}",
                ) from e
        else:
            # get the wrapped embeddings instance
            try:
                return EmbeddingsRegistry.wrap(model, **kwargs)
            except Exception as e:
                raise ValueError(f"Failed to wrap embeddings instance: {e}") from e



"""Comprehensive tests for the chonkie viz module."""

import html
import logging
import os
from pathlib import Path
from unittest.mock import MagicMock, patch

import pytest

from chonkie.types import Chunk
from chonkie.utils.viz import (
    BODY_BACKGROUND_COLOR_DARK,
    BODY_BACKGROUND_COLOR_LIGHT,
    CONTENT_BACKGROUND_COLOR_DARK,
    CONTENT_BACKGROUND_COLOR_LIGHT,
    DARK_THEMES,
    FOOTER_TEMPLATE,
    HTML_TEMPLATE,
    LIGHT_THEMES,
    MAIN_TEMPLATE,
    TEXT_COLOR_DARK,
    TEXT_COLOR_LIGHT,
    Visualizer,
)


@pytest.fixture
def sample_chunks() -> list[Chunk]:
    """Create sample chunks for testing."""
    return [
        Chunk(text="Hello ", start_index=0, end_index=6, token_count=2),
        Chunk(text="world! ", start_index=6, end_index=13, token_count=2),
        Chunk(text="This is ", start_index=13, end_index=21, token_count=3),
        Chunk(text="a test.", start_index=21, end_index=28, token_count=3),
    ]


@pytest.fixture
def sample_text() -> str:
    """Sample text that corresponds to the sample chunks."""
    return "Hello world! This is a test."


@pytest.fixture
def overlapping_chunks() -> list[Chunk]:
    """Create overlapping chunks for testing."""
    return [
        Chunk(text="Hello world", start_index=0, end_index=11, token_count=2),
        Chunk(text="world! This", start_index=6, end_index=17, token_count=3),
        Chunk(text="This is a test", start_index=13, end_index=27, token_count=4),
    ]


@pytest.fixture
def empty_chunks() -> list[Chunk]:
    """Create empty chunks list for testing."""
    return []


@pytest.fixture()
def mock_console_instance() -> MagicMock:
    from rich.console import Console

    return MagicMock(spec=Console)


@pytest.fixture(autouse=True)
def mock_console_class(mock_console_instance) -> MagicMock:
    with patch("rich.console.Console") as mock_console_class:
        mock_console_class.return_value = mock_console_instance
        yield mock_console_class


@pytest.fixture()
def mock_text_instance() -> MagicMock:
    from rich.text import Text

    return MagicMock(spec=Text)


@pytest.fixture()
def mock_text_class(mock_text_instance) -> MagicMock:
    with patch("rich.text.Text") as mock_text_class:
        mock_text_class.return_value = mock_text_instance
        yield mock_text_class


class TestVisualizerInitialization:
    """Test Visualizer initialization and theme handling."""

    def test_init_default_theme(self, mock_console_class) -> None:
        """Test initialization with default theme."""
        viz = Visualizer()
        assert viz.theme_name == "pastel"
        assert viz.theme == LIGHT_THEMES["pastel"]
        assert viz.text_color == TEXT_COLOR_LIGHT
        mock_console_class.assert_called_once()

    def test_init_custom_list_theme(self) -> None:
        """Test initialization with custom list theme."""
        custom_colors = ["#FF0000", "#00FF00", "#0000FF"]
        viz = Visualizer(theme=custom_colors)
        assert viz.theme_name == "custom"
        assert viz.theme == custom_colors
        assert viz.text_color == ""

    def test_init_light_theme(self) -> None:
        """Test initialization with light theme."""
        viz = Visualizer(theme="tiktokenizer")
        assert viz.theme_name == "tiktokenizer"
        assert viz.theme == LIGHT_THEMES["tiktokenizer"]
        assert viz.text_color == TEXT_COLOR_LIGHT

    def test_init_dark_theme(self) -> None:
        """Test initialization with dark theme."""
        viz = Visualizer(theme="tiktokenizer_dark")
        assert viz.theme_name == "tiktokenizer_dark"
        assert viz.theme == DARK_THEMES["tiktokenizer_dark"]
        assert viz.text_color == TEXT_COLOR_DARK

    def test_init_ocean_breeze_theme(self) -> None:
        """Test initialization with the new 'ocean_breeze' light theme."""
        viz = Visualizer(theme="ocean_breeze")
        assert viz.theme_name == "ocean_breeze"
        assert viz.theme == LIGHT_THEMES["ocean_breeze"]
        assert viz.text_color == TEXT_COLOR_LIGHT

    def test_init_midnight_theme(self) -> None:
        """Test initialization with the new 'midnight' dark theme."""
        viz = Visualizer(theme="midnight")
        assert viz.theme_name == "midnight"
        assert viz.theme == DARK_THEMES["midnight"]
        assert viz.text_color == TEXT_COLOR_DARK

    def test_init_invalid_theme(self) -> None:
        """Test initialization with invalid theme."""
        with pytest.raises(ValueError, match="Invalid theme"):
            Visualizer(theme="invalid_theme")


class TestVisualizerThemeManagement:
    """Test theme management methods."""

    def test_get_theme_light(self) -> None:
        """Test getting light theme."""
        viz = Visualizer()
        theme, text_color = viz._get_theme("pastel")
        assert theme == LIGHT_THEMES["pastel"]
        assert text_color == TEXT_COLOR_LIGHT

    def test_get_theme_dark(self) -> None:
        """Test getting dark theme."""
        viz = Visualizer()
        theme, text_color = viz._get_theme("pastel_dark")
        assert theme == DARK_THEMES["pastel_dark"]
        assert text_color == TEXT_COLOR_DARK

    def test_get_color_cycling(self) -> None:
        """Test color cycling behavior."""
        viz = Visualizer(theme="pastel")
        assert viz._get_color(0) == LIGHT_THEMES["pastel"][0]
        assert viz._get_color(1) == LIGHT_THEMES["pastel"][1]
        theme_length = len(LIGHT_THEMES["pastel"])
        assert viz._get_color(theme_length) == LIGHT_THEMES["pastel"][0]
        assert viz._get_color(theme_length + 1) == LIGHT_THEMES["pastel"][1]

    def test_darken_color_valid_hex(self) -> None:
        """Test color darkening with valid hex colors."""
        viz = Visualizer()
        darkened = viz._darken_color("#FF0000", 0.5)
        assert darkened == "#7f0000"
        darkened = viz._darken_color("#F00", 0.5)
        assert darkened == "#7f0000"

    def test_darken_color_invalid_hex(self, caplog) -> None:
        """Test color darkening with invalid hex colors."""
        viz = Visualizer()
        result = viz._darken_color("invalid", 0.5)
        assert result == "#808080"
        assert "Could not darken" in caplog.text

    def test_darken_color_edge_cases(self) -> None:
        """Test color darkening edge cases."""
        viz = Visualizer()
        darkened = viz._darken_color("#FFFFFF", 0.0)
        assert darkened == "#000000"
        darkened = viz._darken_color("#FFFFFF", 1.0)
        assert darkened == "#ffffff"


class TestVisualizerPrintMethod:
    """Test the print method functionality."""

    def test_print_empty_chunks(self, sample_text: str, mock_console_instance) -> None:
        """Test printing empty chunks."""
        Visualizer().print([])
        mock_console_instance.print.assert_called_once_with("No chunks to visualize.")

    @pytest.mark.parametrize("method", ["print", "__call__"])
    def test_print_with_full_text(
        self,
        sample_chunks: list[Chunk],
        sample_text: str,
        mock_console_instance,
        mock_text_class,
        mock_text_instance,
        method: str,
    ) -> None:
        """Test printing with provided full text."""
        viz = Visualizer()
        getattr(viz, method)(sample_chunks, sample_text)
        mock_console_instance.print.assert_called_once_with(mock_text_instance)

    @pytest.mark.parametrize("method", ["print", "__call__"])
    def test_print_without_full_text(
        self,
        sample_chunks: list[Chunk],
        mock_console_instance,
        mock_text_class,
        method: str,
    ) -> None:
        """Test printing without provided full text (reconstruction)."""
        viz = Visualizer()
        getattr(viz, method)(sample_chunks)
        expected_text = "Hello world! This is a test."
        mock_text_class.assert_called_once_with(expected_text)

    def test_print_invalid_chunks_no_text_attr(self) -> None:
        """Test printing with chunks missing text attribute."""
        with pytest.raises(ValueError, match="Chunks must have 'text'"):
            Visualizer().print([[]])

    def test_print_invalid_chunk_indices(self, sample_text: str, caplog) -> None:
        """Test printing with chunks having invalid indices."""
        invalid_chunks = [
            MagicMock(text="Hello ", start_index=None, end_index=6),
            MagicMock(text="world!", start_index=6, end_index="invalid"),
        ]
        Visualizer().print(invalid_chunks, sample_text)
        assert "invalid start/end index" in caplog.text

    def test_print_stylize_error_handling(
        self,
        sample_chunks: list[Chunk],
        sample_text: str,
        mock_text_class,
        mock_text_instance,
        caplog,
    ) -> None:
        """Test error handling during text stylization."""
        mock_text_instance.stylize.side_effect = Exception("Stylize error")
        Visualizer().print(sample_chunks, sample_text)
        assert "Could not apply style" in caplog.text


class TestVisualizerSaveMethod:
    """Test the save method functionality."""

    def test_save_empty_chunks(self, caplog) -> None:
        """Test saving with empty chunks."""
        caplog.set_level(logging.INFO)
        Visualizer().save("test.html", [])
        assert "No chunks to visualize. HTML file not saved." in caplog.text

    def test_save_with_full_text(
        self, sample_chunks: list[Chunk], sample_text: str, tmp_path, caplog
    ) -> None:
        """Test saving with provided full text."""
        caplog.set_level(logging.INFO)
        filepath = os.path.join(tmp_path, "test.html")
        Visualizer().save(filepath, sample_chunks, sample_text)
        with open(filepath, "r", encoding="utf-8") as f:
            content = f.read()
        assert "Hello" in content
        assert "world!" in content
        assert "This is" in content
        assert "a test." in content
        assert "Chunk Visualization" in content
        assert "background-color:" in content
        # Check that logger.info was called with the success message
        assert f"file://{os.path.abspath(filepath)}" in caplog.text

    def test_save_without_full_text(self, sample_chunks: list[Chunk], tmp_path: Path) -> None:
        """Test saving without provided full text (reconstruction)."""
        filepath = os.path.join(tmp_path, "test.html")
        Visualizer().save(filepath, sample_chunks)
        with open(filepath, "r", encoding="utf-8") as f:
            content = f.read()
        assert "Hello" in content
        assert "world!" in content
        assert "This is" in content
        assert "a test." in content

    def test_save_auto_add_html_extension(
        self, sample_chunks: list[Chunk], tmp_path: Path
    ) -> None:
        """Test automatic addition of .html extension."""
        filepath = os.path.join(tmp_path, "test")
        Visualizer().save(filepath, sample_chunks)
        expected_path = filepath + ".html"
        assert os.path.exists(expected_path)

    def test_save_custom_title(self, sample_chunks: list[Chunk], tmp_path: Path) -> None:
        """Test saving with custom title."""
        filepath = os.path.join(tmp_path, "test.html")
        custom_title = "My Custom Title"
        Visualizer().save(filepath, sample_chunks, title=custom_title)
        with open(filepath, "r", encoding="utf-8") as f:
            content = f.read()
        assert f"<title>{html.escape(custom_title)}</title>" in content

    def test_save_dark_theme_styling(self, sample_chunks: list[Chunk], tmp_path: Path) -> None:
        """Test saving with dark theme styling."""
        filepath = os.path.join(tmp_path, "test.html")
        viz = Visualizer(theme="pastel_dark")
        viz.save(filepath, sample_chunks)
        with open(filepath, "r", encoding="utf-8") as f:
            content = f.read()
        assert BODY_BACKGROUND_COLOR_DARK in content
        assert CONTENT_BACKGROUND_COLOR_DARK in content
        assert TEXT_COLOR_DARK in content

    def test_save_light_theme_styling(self, sample_chunks: list[Chunk], tmp_path: Path) -> None:
        """Test saving with light theme styling."""
        filepath = os.path.join(tmp_path, "test.html")
        viz = Visualizer(theme="pastel")
        viz.save(filepath, sample_chunks)
        with open(filepath, "r", encoding="utf-8") as f:
            content = f.read()
        assert BODY_BACKGROUND_COLOR_LIGHT in content
        assert CONTENT_BACKGROUND_COLOR_LIGHT in content
        assert TEXT_COLOR_LIGHT in content

    def test_save_overlapping_chunks(
        self, overlapping_chunks: list[Chunk], tmp_path: Path
    ) -> None:
        """Test saving with overlapping chunks."""
        filepath = os.path.join(tmp_path, "test.html")
        full_text = "Hello world! This is a test."
        Visualizer().save(filepath, overlapping_chunks, full_text)
        with open(filepath, "r", encoding="utf-8") as f:
            content = f.read()
        assert "(Overlap)" in content
        assert "title=" in content

    def test_save_hippo_favicon_embedding(
        self, sample_chunks: list[Chunk], tmp_path: Path
    ) -> None:
        """Test that hippo favicon is properly embedded."""
        filepath = os.path.join(tmp_path, "test.html")
        Visualizer().save(filepath, sample_chunks)
        with open(filepath, "r", encoding="utf-8") as f:
            content = f.read()
        assert 'rel="icon"' in content
        assert 'type="image/svg+xml"' in content
        assert "data:image/svg+xml;base64," in content

    def test_save_favicon_encoding_error(
        self, sample_chunks: list[Chunk], tmp_path, caplog
    ) -> None:
        """Test handling of favicon encoding errors."""
        with patch("base64.b64encode", side_effect=Exception("Encoding error")):
            filepath = os.path.join(tmp_path, "test.html")
            Visualizer().save(filepath, sample_chunks)
            assert "Could not encode embedded hippo favicon: Encoding error" in caplog.text

    def test_save_invalid_chunks(self) -> None:
        """Test saving with invalid chunks."""
        with pytest.raises(AttributeError, match="Chunks must have 'text'"):
            Visualizer().save("test.html", [[]])

    def test_save_file_write_error(self, sample_chunks: list[Chunk]) -> None:
        """Test handling of file write errors."""
        with pytest.raises(IOError, match="Could not write file"):
            Visualizer().save("/nonexistent/directory/test.html", sample_chunks)

    def test_save_text_escaping(self, tmp_path: Path) -> None:
        """Test proper HTML escaping of text content."""
        html_chunks = [
            Chunk(text="<script>", start_index=0, end_index=8, token_count=1),
            Chunk(text="alert('hi')", start_index=8, end_index=19, token_count=2),
            Chunk(text="</script>", start_index=19, end_index=28, token_count=1),
        ]
        html_text = "<script>alert('hi')</script>"
        filepath = os.path.join(tmp_path, "test.html")
        Visualizer().save(filepath, html_chunks, html_text)
        with open(filepath, "r", encoding="utf-8") as f:
            content = f.read()
        assert "&lt;script&gt;" in content
        assert "&lt;/script&gt;" in content
        assert "alert(&#x27;hi&#x27;)" in content

    def test_save_newline_handling(self, tmp_path: Path) -> None:
        """Test proper handling of newlines in text."""
        newline_chunks = [
            Chunk(text="Line 1\n", start_index=0, end_index=7, token_count=2),
            Chunk(text="Line 2\n", start_index=7, end_index=14, token_count=2),
            Chunk(text="Line 3", start_index=14, end_index=20, token_count=2),
        ]
        newline_text = "Line 1\nLine 2\nLine 3"
        filepath = os.path.join(tmp_path, "test.html")
        Visualizer().save(filepath, newline_chunks, newline_text)
        with open(filepath, "r", encoding="utf-8") as f:
            content = f.read()
        assert "Line 1<br>" in content
        assert "Line 2<br>" in content


class TestVisualizerUtilityMethods:
    """Test utility methods."""

    def test_repr(self) -> None:
        """Test __repr__ method."""
        viz = Visualizer(theme="pastel")
        repr_str = repr(viz)
        assert "Visualizer" in repr_str
        assert str(viz.theme) in repr_str


class TestVisualizerEdgeCases:
    """Test edge cases and boundary conditions."""

    def test_empty_text_chunks(self, tmp_path: Path) -> None:
        """Test chunks with empty text."""
        empty_text_chunks = [
            Chunk(text="", start_index=0, end_index=0, token_count=0),
            Chunk(text="Hello", start_index=0, end_index=5, token_count=1),
        ]
        filepath = os.path.join(tmp_path, "test.html")
        Visualizer().save(filepath, empty_text_chunks, "Hello")
        assert os.path.exists(filepath)

    def test_chunks_beyond_text_length(self, sample_text: str, mock_text_class) -> None:
        """Test chunks with indices beyond text length."""
        beyond_chunks = [
            Chunk(text="Hello", start_index=0, end_index=5, token_count=1),
            Chunk(text="Beyond", start_index=100, end_index=106, token_count=1),
        ]
        Visualizer().print(beyond_chunks, sample_text)
        mock_text_class.assert_called_once_with(sample_text)

    def test_negative_chunk_indices(self, sample_text: str, tmp_path: Path) -> None:
        """Test chunks with negative indices."""
        negative_chunks = [
            Chunk(text="Invalid", start_index=-5, end_index=5, token_count=1),
            Chunk(text="Hello", start_index=0, end_index=5, token_count=1),
        ]
        filepath = os.path.join(tmp_path, "test.html")
        Visualizer().save(filepath, negative_chunks, sample_text)
        assert os.path.exists(filepath)

    def test_chunk_start_equals_end(self, sample_text: str) -> None:
        """Test chunks where start_index equals end_index."""
        equal_indices_chunks = [
            Chunk(text="", start_index=5, end_index=5, token_count=0),
            Chunk(text="Hello", start_index=0, end_index=5, token_count=1),
        ]
        Visualizer().print(equal_indices_chunks, sample_text)


class TestVisualizerConstants:
    """Test module constants and templates."""

    def test_theme_constants(self) -> None:
        """Test that theme constants are properly defined."""
        assert isinstance(LIGHT_THEMES, dict)
        assert isinstance(DARK_THEMES, dict)
        assert "pastel" in LIGHT_THEMES
        assert "tiktokenizer" in LIGHT_THEMES
        assert "pastel_dark" in DARK_THEMES
        assert "tiktokenizer_dark" in DARK_THEMES
        assert "ocean_breeze" in LIGHT_THEMES
        assert "midnight" in DARK_THEMES

    def test_color_constants(self) -> None:
        """Test that color constants are properly defined."""
        assert BODY_BACKGROUND_COLOR_LIGHT.startswith("#")
        assert CONTENT_BACKGROUND_COLOR_LIGHT.startswith("#")
        assert TEXT_COLOR_LIGHT.startswith("#")
        assert BODY_BACKGROUND_COLOR_DARK.startswith("#")
        assert CONTENT_BACKGROUND_COLOR_DARK.startswith("#")
        assert TEXT_COLOR_DARK.startswith("#")

    def test_template_constants(self) -> None:
        """Test that template constants are properly defined."""
        assert isinstance(HTML_TEMPLATE, str)
        assert isinstance(MAIN_TEMPLATE, str)
        assert isinstance(FOOTER_TEMPLATE, str)
        assert "{title}" in HTML_TEMPLATE
        assert "{html_parts}" in MAIN_TEMPLATE
        assert "Chonkie" in FOOTER_TEMPLATE

    def test_hippo_svg_content(self) -> None:
        """Test that hippo SVG content is properly defined."""
        assert "🦛" in Visualizer.HIPPO_SVG_CONTENT
        assert "<svg" in Visualizer.HIPPO_SVG_CONTENT
        assert "</svg>" in Visualizer.HIPPO_SVG_CONTENT


class TestVisualizerIntegration:
    """Integration tests combining multiple features."""

    def test_full_workflow_light_theme(
        self,
        sample_chunks: list[Chunk],
        sample_text: str,
        tmp_path: Path,
    ) -> None:
        """Test complete workflow with light theme."""
        filepath = os.path.join(tmp_path, "integration_test.html")
        viz = Visualizer(theme="pastel")
        viz.save(filepath, sample_chunks, sample_text, title="Integration Test")
        assert os.path.exists(filepath)
        with open(filepath, "r", encoding="utf-8") as f:
            content = f.read()
        assert "Integration Test" in content
        assert "Hello" in content and "world!" in content
        assert "background-color:" in content
        assert "🦛" in content or "data:image/svg+xml" in content
        assert BODY_BACKGROUND_COLOR_LIGHT in content

    def test_full_workflow_dark_theme(
        self,
        sample_chunks: list[Chunk],
        sample_text: str,
        tmp_path: Path,
    ) -> None:
        """Test complete workflow with dark theme."""
        filepath = os.path.join(tmp_path, "integration_test_dark.html")
        viz = Visualizer(theme="pastel_dark")
        viz.save(filepath, sample_chunks, sample_text, title="Dark Theme Test")
        with open(filepath, "r", encoding="utf-8") as f:
            content = f.read()
        assert BODY_BACKGROUND_COLOR_DARK in content
        assert CONTENT_BACKGROUND_COLOR_DARK in content
        assert TEXT_COLOR_DARK in content

    def test_complex_overlapping_scenario(self, tmp_path: Path) -> None:
        """Test complex overlapping chunks scenario."""
        complex_chunks = [
            Chunk(text="The quick", start_index=0, end_index=9, token_count=2),
            Chunk(text="quick brown", start_index=4, end_index=15, token_count=2),
            Chunk(text="brown fox", start_index=10, end_index=19, token_count=2),
            Chunk(text="fox jumps", start_index=16, end_index=25, token_count=2),
        ]
        complex_text = "The quick brown fox jumps"
        filepath = os.path.join(tmp_path, "complex_test.html")
        Visualizer().save(filepath, complex_chunks, complex_text)
        with open(filepath, "r", encoding="utf-8") as f:
            content = f.read()
        assert "(Overlap)" in content
        assert "The" in content
        assert "quick" in content
        assert "brown" in content
        assert "fox" in content
        assert "jumps" in content


class TestVisualizerStringInput:
    """Test the list[str] input support for Visualizer."""

    def test_print_list_of_strings(
        self,
        mock_console_instance,
        mock_text_class,
        mock_text_instance,
    ) -> None:
        """Test printing a list of strings auto-constructs chunks and full_text."""
        Visualizer().print(["hello", "world"])
        mock_text_class.assert_called_once_with("helloworld")
        mock_console_instance.print.assert_called_once_with(mock_text_instance)

    def test_call_list_of_strings(self, mock_text_class, mock_console_class) -> None:
        """Test __call__ with a list of strings delegates to print correctly."""
        Visualizer()(["hello", "world"])
        mock_text_class.assert_called_once_with("helloworld")

    def test_print_strings_with_explicit_full_text(self, mock_text_class) -> None:
        """Test that explicit full_text is not overwritten when strings are passed."""
        # full_text has a space between the chunks
        Visualizer().print(["hello", "world"], full_text="hello world")
        mock_text_class.assert_called_once_with("hello world")

    def test_print_strings_indices_with_gaps(self, mock_text_class, mock_text_instance) -> None:
        """Test that string chunks get correct indices when full_text has gaps."""
        Visualizer().print(["a", "b"], full_text="a-b")
        # "b" should be highlighted at index 2, not index 1
        assert mock_text_instance.stylize.call_count == 2
        # First call: "a" at (0, 1)
        first_call_args = mock_text_instance.stylize.call_args_list[0]
        assert first_call_args[0][1] == 0  # start
        assert first_call_args[0][2] == 1  # end
        # Second call: "b" at (2, 3)
        second_call_args = mock_text_instance.stylize.call_args_list[1]
        assert second_call_args[0][1] == 2  # start
        assert second_call_args[0][2] == 3  # end

    def test_print_mixed_chunks_and_strings(self, mock_text_class, mock_text_instance) -> None:
        """Test printing a mix of Chunk objects and strings."""
        mixed = [
            Chunk(text="Hello ", start_index=0, end_index=6, token_count=1),
            "world",
        ]
        Visualizer().print(mixed, full_text="Hello world")
        mock_text_class.assert_called_once_with("Hello world")
        # Should have 2 stylize calls for 2 chunks
        assert mock_text_instance.stylize.call_count == 2



"""Cloud Pipeline for Chonkie API."""

import builtins
import os
import re
from dataclasses import dataclass, field
from typing import Any, Optional, Union

import httpx

from chonkie.types import Chunk

from .file import FileManager

# Slug validation pattern: lowercase letters, numbers, dashes, underscores
SLUG_PATTERN = re.compile(r"^[a-z0-9_-]+$")


@dataclass
class PipelineStep:
    """A single step in a pipeline configuration."""

    type: str
    component: str
    params: dict[str, Any] = field(default_factory=dict)

    def to_dict(self) -> dict[str, Any]:
        """Convert to dictionary for API payload."""
        return {
            "type": self.type,
            "component": self.component,
            **self.params,
        }

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> "PipelineStep":
        """Create from dictionary."""
        step_type = data.get("type", "")
        component = data.get("component", "")
        params = {k: v for k, v in data.items() if k not in ["type", "component"]}
        return cls(type=step_type, component=component, params=params)


class Pipeline:
    """Cloud Pipeline - build and execute pipelines via Chonkie API.

    Examples:
        ```python
        from chonkie.cloud import Pipeline

        # Create new pipeline
        pipeline = Pipeline(
            slug="my-rag-pipeline",
            description="Pipeline for RAG processing"
        ).chunk_with("recursive", chunk_size=512).refine_with("overlap", context_size=64)

        # Execute with text
        chunks = pipeline.run(text="Your document text here")

        # Execute with file (auto-uploaded)
        chunks = pipeline.run(file="document.pdf")

        # Fetch existing pipeline
        pipeline = Pipeline.get("my-rag-pipeline")
        chunks = pipeline.run(text="...")

        # List all pipelines
        for p in Pipeline.list():
            print(f"{p.slug}: {p.describe()}")
        ```

    """

    BASE_URL = "https://api.chonkie.ai"
    VERSION = "v1"

    def __init__(
        self,
        slug: str,
        description: Optional[str] = None,
        api_key: Optional[str] = None,
    ) -> None:
        """Initialize a new cloud Pipeline.

        Args:
            slug: Unique identifier for the pipeline. Must contain only lowercase
                  letters, numbers, dashes, and underscores.
            description: Optional description of the pipeline.
            api_key: Chonkie API key. If None, reads from CHONKIE_API_KEY env var.

        Raises:
            ValueError: If slug format is invalid or API key is not provided.

        """
        # Validate API key
        self._api_key = api_key or os.getenv("CHONKIE_API_KEY")
        if not self._api_key:
            raise ValueError(
                "No API key provided. Please set the CHONKIE_API_KEY environment "
                "variable or pass an api_key to the Pipeline constructor.",
            )

        # Validate slug format
        if not SLUG_PATTERN.match(slug):
            raise ValueError(
                f"Invalid slug '{slug}'. Slug must contain only lowercase letters, "
                "numbers, dashes, and underscores.",
            )

        self._slug = slug
        self._description = description
        self._steps: list[PipelineStep] = []
        self._is_saved = False
        self._id: Optional[str] = None
        self._created_at: Optional[str] = None
        self._updated_at: Optional[str] = None

        # Initialize file manager for file uploads
        self._file_manager = FileManager(api_key=self._api_key)

    @property
    def slug(self) -> str:
        """Return the pipeline slug."""
        return self._slug

    @property
    def description(self) -> Optional[str]:
        """Return the pipeline description."""
        return self._description

    @property
    def is_saved(self) -> bool:
        """Return True if pipeline exists in cloud."""
        return self._is_saved

    @property
    def steps(self) -> list[PipelineStep]:
        """Return the pipeline steps."""
        return self._steps.copy()

    def _get_headers(self) -> dict[str, str]:
        """Get headers for API requests."""
        return {
            "Authorization": f"Bearer {self._api_key}",
            "Content-Type": "application/json",
        }

    @classmethod
    def get(cls, slug: str, api_key: Optional[str] = None) -> "Pipeline":
        """Fetch an existing pipeline from the cloud.

        Args:
            slug: The pipeline slug to fetch.
            api_key: Chonkie API key. If None, reads from CHONKIE_API_KEY env var.

        Returns:
            Pipeline instance with the fetched configuration.

        Raises:
            ValueError: If pipeline not found or API error occurs.

        """
        api_key = api_key or os.getenv("CHONKIE_API_KEY")
        if not api_key:
            raise ValueError(
                "No API key provided. Please set the CHONKIE_API_KEY environment "
                "variable or pass an api_key.",
            )

        response = httpx.get(
            f"{cls.BASE_URL}/{cls.VERSION}/pipeline/{slug}",
            headers={
                "Authorization": f"Bearer {api_key}",
                "Content-Type": "application/json",
            },
        )

        if response.status_code == 404:
            raise ValueError(f"Pipeline '{slug}' not found.")
        if response.status_code != 200:
            raise ValueError(f"Failed to fetch pipeline: {response.text}")

        data = response.json()

        # Create pipeline instance
        pipeline = cls(
            slug=data["slug"],
            description=data.get("description"),
            api_key=api_key,
        )

        # Populate from response
        pipeline._is_saved = True
        pipeline._id = data.get("id")
        pipeline._created_at = data.get("created_at")
        pipeline._updated_at = data.get("updated_at")
        pipeline._steps = [PipelineStep.from_dict(step) for step in data.get("steps", [])]

        return pipeline

    @classmethod
    def list(cls, api_key: Optional[str] = None) -> list["Pipeline"]:
        """List all pipelines from the cloud.

        Args:
            api_key: Chonkie API key. If None, reads from CHONKIE_API_KEY env var.

        Returns:
            List of Pipeline instances.

        Raises:
            ValueError: If API error occurs.

        """
        api_key = api_key or os.getenv("CHONKIE_API_KEY")
        if not api_key:
            raise ValueError(
                "No API key provided. Please set the CHONKIE_API_KEY environment "
                "variable or pass an api_key.",
            )

        response = httpx.get(
            f"{cls.BASE_URL}/{cls.VERSION}/pipeline",
            headers={
                "Authorization": f"Bearer {api_key}",
                "Content-Type": "application/json",
            },
        )

        if response.status_code != 200:
            raise ValueError(f"Failed to list pipelines: {response.text}")

        data = response.json()
        pipelines = []

        for p in data.get("pipelines", []):
            pipeline = cls(
                slug=p["slug"],
                description=p.get("description"),
                api_key=api_key,
            )
            pipeline._is_saved = True
            pipeline._id = p.get("id")
            pipeline._created_at = p.get("created_at")
            pipeline._updated_at = p.get("updated_at")
            pipeline._steps = [PipelineStep.from_dict(step) for step in p.get("steps", [])]
            pipelines.append(pipeline)

        return pipelines

    @classmethod
    def validate(
        cls,
        steps: builtins.list[dict[str, Any]],
        api_key: Optional[str] = None,
    ) -> tuple[bool, Optional[builtins.list[str]]]:
        """Validate a pipeline configuration via the cloud API.

        Args:
            steps: List of step configurations to validate.
            api_key: Chonkie API key. If None, reads from CHONKIE_API_KEY env var.

        Returns:
            Tuple of (is_valid, errors). errors is None if valid.

        Raises:
            ValueError: If API error occurs.

        """
        api_key = api_key or os.getenv("CHONKIE_API_KEY")
        if not api_key:
            raise ValueError(
                "No API key provided. Please set the CHONKIE_API_KEY environment "
                "variable or pass an api_key.",
            )

        # Convert to API format
        formatted_steps = []
        for step in steps:
            if isinstance(step, PipelineStep):
                formatted_steps.append(step.to_dict())
            elif isinstance(step, dict):
                formatted_steps.append(step)
            else:
                raise ValueError(f"Invalid step format: {type(step)}")

        response = httpx.post(
            f"{cls.BASE_URL}/{cls.VERSION}/pipeline/validate",
            json={"steps": formatted_steps},
            headers={
                "Authorization": f"Bearer {api_key}",
                "Content-Type": "application/json",
            },
        )

        if response.status_code != 200:
            raise ValueError(f"Validation request failed: {response.text}")

        data = response.json()
        return data.get("valid", False), data.get("errors")

    def chunk_with(self, chunker_type: str, **kwargs: Any) -> "Pipeline":
        """Add a chunking step to the pipeline.

        Args:
            chunker_type: Type of chunker (e.g., "recursive", "semantic", "sentence").
            **kwargs: Chunker parameters (e.g., chunk_size=512).

        Returns:
            Pipeline instance for method chaining.

        Example:
            ```python
            pipeline.chunk_with("recursive", chunk_size=512, tokenizer="gpt2")
            ```

        """
        self._steps.append(PipelineStep(type="chunk", component=chunker_type, params=kwargs))
        return self

    def refine_with(self, refinery_type: str, **kwargs: Any) -> "Pipeline":
        """Add a refinement step to the pipeline.

        Args:
            refinery_type: Type of refinery (e.g., "overlap", "embeddings").
            **kwargs: Refinery parameters.

        Returns:
            Pipeline instance for method chaining.

        Example:
            ```python
            pipeline.refine_with("overlap", context_size=64)
            pipeline.refine_with("embeddings", embedding_model="text-embedding-3-small")
            ```

        """
        self._steps.append(PipelineStep(type="refine", component=refinery_type, params=kwargs))
        return self

    def process_with(self, chef_type: str, **kwargs: Any) -> "Pipeline":
        """Add a processing step to the pipeline.

        Args:
            chef_type: Type of chef/processor (e.g., "markdown", "text").
            **kwargs: Processor parameters.

        Returns:
            Pipeline instance for method chaining.

        Example:
            ```python
            pipeline.process_with("markdown")
            ```

        """
        self._steps.append(PipelineStep(type="process", component=chef_type, params=kwargs))
        return self

    def _save(self) -> "Pipeline":
        """Save the pipeline to the cloud (create new).

        Returns:
            Pipeline instance for method chaining.

        Raises:
            ValueError: If pipeline already exists or API error occurs.

        """
        if not self._steps:
            raise ValueError("Cannot save pipeline with no steps.")

        payload = {
            "slug": self._slug,
            "description": self._description,
            "steps": [step.to_dict() for step in self._steps],
        }

        response = httpx.post(
            f"{self.BASE_URL}/{self.VERSION}/pipeline",
            json=payload,
            headers=self._get_headers(),
        )

        if response.status_code != 200:
            # Pipeline might already exist, try to update instead
            return self.update()

        data = response.json()
        self._is_saved = True
        self._id = data.get("id")
        self._created_at = data.get("created_at")
        self._updated_at = data.get("updated_at")

        return self

    def update(self, description: Optional[str] = None) -> "Pipeline":
        """Update the pipeline in the cloud.

        Args:
            description: New description (optional).

        Returns:
            Pipeline instance for method chaining.

        Raises:
            ValueError: If pipeline doesn't exist or API error occurs.

        """
        payload: dict[str, Any] = {}

        if description is not None:
            payload["description"] = description
            self._description = description

        if self._steps:
            payload["steps"] = [step.to_dict() for step in self._steps]

        if not payload:
            return self  # Nothing to update

        response = httpx.put(
            f"{self.BASE_URL}/{self.VERSION}/pipeline/{self._slug}",
            json=payload,
            headers=self._get_headers(),
        )

        if response.status_code == 404:
            raise ValueError(f"Pipeline '{self._slug}' not found.")
        if response.status_code != 200:
            raise ValueError(f"Failed to update pipeline: {response.text}")

        data = response.json()
        self._is_saved = True
        self._updated_at = data.get("updated_at")

        return self

    def delete(self) -> None:
        """Delete the pipeline from the cloud.

        Raises:
            ValueError: If pipeline doesn't exist or API error occurs.

        """
        response = httpx.delete(
            f"{self.BASE_URL}/{self.VERSION}/pipeline/{self._slug}",
            headers=self._get_headers(),
        )

        if response.status_code == 404:
            raise ValueError(f"Pipeline '{self._slug}' not found.")
        if response.status_code != 200:
            raise ValueError(f"Failed to delete pipeline: {response.text}")

        self._is_saved = False
        self._id = None

    def run(
        self,
        text: Optional[Union[str, builtins.list[str]]] = None,
        file: str | os.PathLike | None = None,
    ) -> builtins.list[Chunk]:
        """Execute the pipeline via the cloud API.

        Args:
            text: Text to process. Can be a single string or list of strings.
            file: Path to file to process. Will be uploaded automatically.

        Returns:
            List of Chunk objects.

        Raises:
            ValueError: If neither text nor file provided, or API error occurs.

        """
        if text is None and file is None:
            raise ValueError("Either 'text' or 'file' must be provided.")
        if text is not None and file is not None:
            raise ValueError("Cannot provide both 'text' and 'file'.")

        # Save pipeline if not already saved
        if not self._is_saved:
            self._save()

        # Build payload
        payload: dict[str, Any] = {}

        if file is not None:
            # Upload file first
            uploaded = self._file_manager.upload(file)
            payload["file"] = {"type": "document", "content": uploaded.name}
        else:
            payload["text"] = text

        # Execute pipeline
        response = httpx.post(
            f"{self.BASE_URL}/{self.VERSION}/pipeline/{self._slug}",
            json=payload,
            headers=self._get_headers(),
        )

        if response.status_code == 404:
            raise ValueError(f"Pipeline '{self._slug}' not found.")
        if response.status_code != 200:
            raise ValueError(f"Pipeline execution failed: {response.text}")

        data = response.json()
        chunks = [Chunk.from_dict(chunk) for chunk in data.get("chunks", [])]

        return chunks

    def to_config(self) -> builtins.list[dict[str, Any]]:
        """Export pipeline configuration as a list of step dictionaries.

        Returns:
            List of step configurations.

        """
        return [step.to_dict() for step in self._steps]

    def describe(self) -> str:
        """Get a human-readable description of the pipeline.

        Returns:
            String description of the pipeline steps.

        """
        if not self._steps:
            return "Empty pipeline"

        descriptions = []
        for step in self._steps:
            descriptions.append(f"{step.type}({step.component})")

        return " -> ".join(descriptions)

    def reset(self) -> "Pipeline":
        """Reset the pipeline steps.

        Returns:
            Pipeline instance for method chaining.

        """
        self._steps.clear()
        return self

    def __repr__(self) -> str:
        """Return string representation of the pipeline."""
        saved_status = "saved" if self._is_saved else "not saved"
        return f"Pipeline(slug='{self._slug}', {saved_status}, {self.describe()})"



"""Mixedbread embeddings - thin wrapper around CatsuEmbeddings."""

import importlib.util as importutil
import os
from typing import Any, Optional

import numpy as np

from .base import BaseEmbeddings
from .catsu import CatsuEmbeddings


class MixedbreadEmbeddings(BaseEmbeddings):
    """Mixedbread embeddings via CatsuEmbeddings.

    Wraps CatsuEmbeddings with provider="mixedbread" for convenient access to
    Mixedbread's embedding models. Equivalent to:
    CatsuEmbeddings(model=..., provider="mixedbread")

    Args:
        model: Mixedbread embedding model name (default: "mxbai-embed-large-v1").
        api_key: Mixedbread API key (or set MIXEDBREAD_API_KEY env var).
        max_retries: Maximum retry attempts (default: 3).
        timeout: Request timeout in seconds (default: 30).
        batch_size: Number of texts per API call (default: 128).

    """

    DEFAULT_MODEL = "mxbai-embed-large-v1"

    def __init__(
        self,
        model: str = DEFAULT_MODEL,
        api_key: Optional[str] = None,
        max_retries: int = 3,
        timeout: int = 30,
        batch_size: int = 128,
    ):
        """Initialize Mixedbread embeddings.

        Args:
            model: Mixedbread embedding model name.
            api_key: Mixedbread API key (falls back to MIXEDBREAD_API_KEY env var).
            max_retries: Maximum retry attempts.
            timeout: Request timeout in seconds.
            batch_size: Number of texts per API call.

        Raises:
            ImportError: If the catsu package is not installed.

        """
        super().__init__()

        if not self._is_available():
            raise ImportError(
                "One (or more) of the following packages is not available: catsu. "
                'Please install it via `pip install "chonkie[catsu]"`',
            )

        self.model = model
        api_key = api_key or os.getenv("MIXEDBREAD_API_KEY")
        api_keys = {"mixedbread": api_key} if api_key else None

        self._catsu = CatsuEmbeddings(
            model=model,
            provider="mixedbread",
            api_keys=api_keys,
            max_retries=max_retries,
            timeout=timeout,
            batch_size=batch_size,
        )

    def embed(self, text: str) -> np.ndarray:
        """Embed a single text string."""
        return self._catsu.embed(text)

    def embed_batch(self, texts: list) -> list:
        """Embed multiple texts using batched API calls."""
        return self._catsu.embed_batch(texts)

    @property
    def dimension(self) -> int:
        """Return the embedding dimension."""
        return self._catsu.dimension

    def get_tokenizer(self) -> Any:
        """Return a tokenizer object for token counting."""
        return self._catsu.get_tokenizer()

    @classmethod
    def _is_available(cls) -> bool:
        """Check if the catsu package is available."""
        return importutil.find_spec("catsu") is not None

    def __repr__(self) -> str:
        """Return a string representation of the MixedbreadEmbeddings instance."""
        return f"MixedbreadEmbeddings(model={self.model})"



"""Google Gemini embeddings - backward-compatible wrapper around CatsuEmbeddings.

Note: Consider using CatsuEmbeddings(model=..., provider="gemini") directly.
"""

import importlib.util as importutil
import os
import warnings
from typing import Any, Optional

import numpy as np

from .base import BaseEmbeddings
from .catsu import CatsuEmbeddings


class GeminiEmbeddings(BaseEmbeddings):
    """Google Gemini embeddings via CatsuEmbeddings.

    This is a backward-compatible wrapper around CatsuEmbeddings.
    Consider using CatsuEmbeddings(model=..., provider="gemini") directly.

    Args:
        model: Gemini embedding model name (default: "gemini-embedding-001").
        api_key: Gemini API key (or set GEMINI_API_KEY env var).
        task_type: Ignored; kept for backward compatibility.
        max_retries: Maximum retry attempts (default: 3).
        batch_size: Number of texts per API call (default: 100).
        show_warnings: Ignored; kept for backward compatibility.

    """

    DEFAULT_MODEL = "gemini-embedding-001"

    AVAILABLE_MODELS = {
        "text-embedding-004": (768, 2048),
        "embedding-001": (768, 2048),
        "gemini-embedding-001": (3072, 8192),
    }

    def __init__(
        self,
        model: str = DEFAULT_MODEL,
        api_key: Optional[str] = None,
        task_type: str = "SEMANTIC_SIMILARITY",
        max_retries: int = 3,
        batch_size: int = 100,
        show_warnings: bool = True,
    ):
        """Initialize Gemini embeddings wrapper.

        Args:
            model: Gemini embedding model name.
            api_key: Gemini API key (falls back to GEMINI_API_KEY env var).
            task_type: Ignored; kept for backward compatibility.
            max_retries: Maximum retry attempts.
            batch_size: Number of texts per API call.
            show_warnings: Ignored; kept for backward compatibility.

        Raises:
            ImportError: If the catsu package is not installed.

        """
        super().__init__()

        if not self._is_available():
            raise ImportError(
                "One (or more) of the following packages is not available: catsu. "
                'Please install it via `pip install "chonkie[catsu]"`',
            )

        if task_type != "SEMANTIC_SIMILARITY":
            warnings.warn(
                "The `task_type` parameter is not supported in this version and will be ignored.",
                DeprecationWarning,
                stacklevel=2,
            )
        if show_warnings is not True:
            warnings.warn(
                "The `show_warnings` parameter is not supported in this version and will be ignored.",
                DeprecationWarning,
                stacklevel=2,
            )

        self.model = model if model else self.DEFAULT_MODEL
        self.task_type = task_type
        api_key = api_key or os.getenv("GEMINI_API_KEY")
        api_keys = {"gemini": api_key} if api_key else None

        self._catsu = CatsuEmbeddings(
            model=self.model,
            provider="gemini",
            api_keys=api_keys,
            max_retries=max_retries,
            batch_size=batch_size,
        )

    def embed(self, text: str) -> np.ndarray:
        """Embed a single text string."""
        return self._catsu.embed(text)

    def embed_batch(self, texts: list) -> list:
        """Embed multiple texts using batched API calls."""
        return self._catsu.embed_batch(texts)

    async def aembed(self, text: str) -> np.ndarray:
        """Embed a single text string asynchronously."""
        return await self._catsu.aembed(text)

    async def aembed_batch(self, texts: list) -> list:
        """Embed multiple texts asynchronously using batched API calls."""
        return await self._catsu.aembed_batch(texts)

    @property
    def dimension(self) -> int:
        """Return the embedding dimension."""
        return self._catsu.dimension

    def get_tokenizer(self) -> Any:
        """Return a tokenizer object for token counting."""
        return self._catsu.get_tokenizer()

    @classmethod
    def _is_available(cls) -> bool:
        """Check if the catsu package is available."""
        return importutil.find_spec("catsu") is not None

    def __repr__(self) -> str:
        """Representation of the GeminiEmbeddings instance."""
        return f"GeminiEmbeddings(model={self.model}, task_type={self.task_type})"



"""Module for managing access to the Chonkie hub."""

import json
import os
from functools import cache
from pathlib import Path
from typing import Optional


@cache
def _get_recipe_schema(version: str) -> dict:
    # This is memoized to avoid multiple downloads of the same schema;
    # we wouldn't expect a version to change during the runtime of a process.

    from huggingface_hub import hf_hub_download

    path = hf_hub_download(
        repo_id="chonkie-ai/recipes",
        repo_type="dataset",
        filename=f"{version}.schema.json",
    )
    return dict(json.loads(Path(path).read_bytes()))


class Hubbie:
    """Hubbie is a Huggingface hub manager for Chonkie.

    Methods:
        get_recipe(recipe_name: str, lang: Optional[str] = 'en') -> Optional[Dict]:
            Get a recipe from the hub.
        get_recipe_schema() -> Dict:
            Get the current recipe schema from the hub.

    """

    SCHEMA_VERSION = "v1"

    def __init__(self) -> None:
        """Initialize Hubbie."""
        # Lazy import the dependencies (huggingface_hub)

        # define the path to the recipes
        self.get_recipe_config = {
            "repo": "chonkie-ai/recipes",
            "subfolder": "recipes",
            "repo_type": "dataset",
        }

        # define the path to the pipeline recipes
        self.get_pipeline_recipe_config = {
            "repo": "chonkie-ai/recipes",
            "subfolder": "pipelines",
            "repo_type": "dataset",
        }

    @property
    def recipe_schema(self) -> dict:
        """The current recipe schema, from the hub."""
        return _get_recipe_schema(self.SCHEMA_VERSION).copy()

    def get_recipe_schema(self) -> dict:
        """Get the current recipe schema from the hub."""
        return self.recipe_schema

    def _validate_recipe(self, recipe: dict) -> Optional[bool]:
        """Validate a recipe against the current schema."""
        import jsonschema

        try:
            jsonschema.validate(recipe, self.recipe_schema)
            return True
        except jsonschema.ValidationError as error:
            raise ValueError(
                f"Recipe is invalid. Please check the recipe and try again. Error: {error}",
            ) from error

    def get_recipe(
        self,
        name: Optional[str] = "default",
        lang: Optional[str] = "en",
        path: str | os.PathLike | None = None,
    ) -> dict:
        """Get a recipe from the hub.

        Args:
            name (Optional[str]): The name of the recipe to get.
            lang (Optional[str]): The language of the recipe to get.
            path (Optional[str]): Optionally, provide the path to the recipe.

        Returns:
            Optional[Dict]: The recipe.

        Raises:
            ValueError: If the recipe is not found.
            ValueError: If neither (name, lang) nor path are provided.
            ValueError: If the recipe is invalid.

        """
        # Check if either (name & lang) or path is provided
        if (name is None or lang is None) and path is None:
            raise ValueError("Either (name & lang) or path must be provided.")

        from huggingface_hub import hf_hub_download

        # If path is not provided, download the recipe from the hub
        if path is None and (name is not None and lang is not None):
            try:
                path = hf_hub_download(
                    repo_id=self.get_recipe_config["repo"],
                    repo_type=self.get_recipe_config["repo_type"],
                    subfolder=self.get_recipe_config["subfolder"],
                    filename=f"{name}_{lang}.json",
                )
            except Exception as error:
                raise ValueError(
                    f"Could not download recipe '{name}_{lang}'. Ensure name and lang are correct or provide a valid path. Error: {error}",
                ) from error

        # If we couldn't get the path or download the recipe, raise error
        if path is None:
            raise ValueError(
                f"Could not determine path for recipe '{name}_{lang}'. Ensure name and lang are correct or provide a valid path.",
            )

        # using Pathlib to check if the file exists
        path_obj = Path(path)
        if not path_obj.exists():
            raise ValueError(
                f"Failed to get the file {path} —— please check if this file exists and if the path is correct.",
            )

        # Path exists, now open it and load the recipe
        try:
            with path_obj.open("r") as f:
                recipe = dict(json.loads(f.read()))
        except Exception as error:
            raise ValueError(
                f"Failed to read the file {path} —— please check if the file is valid JSON and if the path is correct. Error: {error}",
            ) from error

        # Validate the recipe with jsonschema
        assert self._validate_recipe(recipe), (
            "Recipe is invalid. Please check the recipe and try again."
        )

        # Return the recipe
        return recipe

    def get_pipeline_recipe(self, name: str, path: str | os.PathLike | None = None) -> dict:
        """Get a pipeline recipe from the hub.

        Args:
            name: The name of the pipeline recipe to get.
            path: Optionally, provide the path to the recipe file.

        Returns:
            Dict: The pipeline recipe with 'steps' key.

        Raises:
            ValueError: If the recipe is not found or invalid.

        """
        # If path is not provided, download the recipe from the hub
        if path is None:
            from huggingface_hub import hf_hub_download

            try:
                path = hf_hub_download(
                    repo_id=self.get_pipeline_recipe_config["repo"],
                    repo_type=self.get_pipeline_recipe_config["repo_type"],
                    subfolder=self.get_pipeline_recipe_config["subfolder"],
                    filename=f"{name}.json",
                )
            except Exception as error:
                raise ValueError(
                    f"Could not download pipeline recipe '{name}'. "
                    f"Ensure name is correct or provide a valid path. Error: {error}",
                ) from error

        # If we couldn't get the path, raise error
        if path is None:
            raise ValueError(
                f"Could not determine path for pipeline recipe '{name}'. "
                f"Ensure name is correct or provide a valid path.",
            )

        # Check if file exists
        path_obj = Path(path)
        if not path_obj.exists():
            raise ValueError(
                f"Failed to get the file {path} — please check if this file exists "
                f"and if the path is correct.",
            )

        # Load the recipe
        try:
            with path_obj.open("r") as f:
                recipe = dict(json.loads(f.read()))
        except Exception as error:
            raise ValueError(
                f"Failed to read the file {path} — please check if the file is valid JSON. "
                f"Error: {error}",
            ) from error

        # Validate it has required fields
        if "steps" not in recipe:
            raise ValueError(f"Pipeline recipe '{name}' is missing 'steps' field.")

        # Optionally validate schema version
        if "schema" in recipe and recipe["schema"] != self.SCHEMA_VERSION:
            raise ValueError(
                f"Pipeline recipe '{name}' has schema version '{recipe['schema']}', "
                f"but expected '{self.SCHEMA_VERSION}'.",
            )

        return recipe



"""Module containing TokenChunker class.

This module provides a TokenChunker class for splitting text into chunks of a specified token size.

"""

from typing import Generator, Sequence, Union

from tqdm import trange

from chonkie.chunker.base import BaseChunker
from chonkie.logger import get_logger
from chonkie.pipeline import chunker
from chonkie.tokenizer import TokenizerProtocol
from chonkie.types import Chunk

logger = get_logger(__name__)


@chunker("token")
class TokenChunker(BaseChunker):
    """Chunker that splits text into chunks of a specified token size.

    Args:
        tokenizer: The tokenizer instance to use for encoding/decoding
        chunk_size: Maximum number of tokens per chunk
        chunk_overlap: Number of tokens to overlap between chunks

    """

    def __init__(
        self,
        tokenizer: Union[str, TokenizerProtocol] = "character",
        chunk_size: int = 2048,
        chunk_overlap: Union[int, float] = 0,
    ) -> None:
        """Initialize the TokenChunker with configuration parameters.

        Args:
            tokenizer: The tokenizer instance to use for encoding/decoding
            chunk_size: Maximum number of tokens per chunk
            chunk_overlap: Number of tokens to overlap between chunks

        Raises:
            ValueError: If chunk_size <= 0 or chunk_overlap >= chunk_size

        """
        super().__init__(tokenizer)
        if chunk_size <= 0:
            raise ValueError("chunk_size must be positive")
        if isinstance(chunk_overlap, int) and chunk_overlap >= chunk_size:
            raise ValueError("chunk_overlap must be less than chunk_size")

        # Assign the values if they make sense
        self.chunk_size = chunk_size
        self.chunk_overlap = (
            chunk_overlap if isinstance(chunk_overlap, int) else int(chunk_overlap * chunk_size)
        )

        self._use_multiprocessing = False

    def _create_chunks(
        self,
        chunk_texts: Sequence[str],
        token_groups: list[list[int]],
        token_counts: list[int],
    ) -> list[Chunk]:
        """Create chunks from a list of texts."""
        # Find the overlap lengths for index calculation
        if self.chunk_overlap > 0:
            # we get the overlap texts, that gives you the start_index for the next chunk
            # if the token group is smaller than the overlap, we just use the whole token group
            overlap_texts = self.tokenizer.decode_batch([
                token_group[-self.chunk_overlap :]
                if (len(token_group) > self.chunk_overlap)
                else token_group
                for token_group in token_groups
            ])
            overlap_lengths = [len(overlap_text) for overlap_text in overlap_texts]
        else:
            overlap_lengths = [0] * len(token_groups)

        # Create the chunks
        chunks = []
        current_index = 0
        for chunk_text, overlap_length, token_count in zip(
            chunk_texts,
            overlap_lengths,
            token_counts,
        ):
            start_index = current_index
            end_index = start_index + len(chunk_text)
            chunks.append(
                Chunk(
                    text=chunk_text,
                    start_index=start_index,
                    end_index=end_index,
                    token_count=token_count,
                ),
            )
            current_index = end_index - overlap_length

        return chunks

    def _token_group_generator(self, tokens: Sequence[int]) -> Generator[list[int], None, None]:
        """Generate chunks from a list of tokens."""
        for start in range(0, len(tokens), self.chunk_size - self.chunk_overlap):
            end = min(start + self.chunk_size, len(tokens))
            yield list(tokens[start:end])
            if end == len(tokens):
                break

    def chunk(self, text: str) -> list[Chunk]:
        """Split text into overlapping chunks of specified token size.

        Args:
            text: Input text to be chunked

        Returns:
            List of Chunk objects containing the chunked text and metadata

        """
        if not text.strip():
            return []

        logger.debug(f"Chunking text of length {len(text)} with chunk_size={self.chunk_size}")

        # Encode full text
        text_tokens = self.tokenizer.encode(text)

        # Calculate token groups and counts
        token_groups = list(self._token_group_generator(text_tokens))
        token_counts = [len(toks) for toks in token_groups]

        # decode the token groups into the chunk texts
        chunk_texts = self.tokenizer.decode_batch(token_groups)

        # Create the chunks from the token groups and token counts
        chunks = self._create_chunks(chunk_texts, token_groups, token_counts)

        logger.info(f"Created {len(chunks)} chunks from {len(text_tokens)} tokens")
        return chunks

    def _process_batch(self, texts: list[str]) -> list[list[Chunk]]:
        """Process a batch of texts."""
        # encode the texts into tokens in a batch
        tokens_list = self.tokenizer.encode_batch(texts)
        result: list = []

        for tokens in tokens_list:
            if not tokens:
                result.append([])
                continue

            # get the token groups
            token_groups = list(self._token_group_generator(tokens))

            # get the token counts
            token_counts = [len(token_group) for token_group in token_groups]

            # decode the token groups into the chunk texts
            chunk_texts = self.tokenizer.decode_batch(token_groups)

            # create the chunks from the token groups and token counts
            chunks = self._create_chunks(chunk_texts, token_groups, token_counts)
            result.append(chunks)

        return result

    def chunk_batch(  # type: ignore[override]
        self,
        texts: list[str],
        batch_size: int = 1,
        show_progress_bar: bool = True,
    ) -> list[list[Chunk]]:
        """Split a batch of texts into their respective chunks.

        Args:
            texts: List of input texts to be chunked
            batch_size: Number of texts to process in a single batch
            show_progress_bar: Whether to show a progress bar

        Returns:
            List of lists of Chunk objects containing the chunked text and metadata

        """
        chunks: list = []
        for i in trange(
            0,
            len(texts),
            batch_size,
            desc="🦛",
            disable=not show_progress_bar,
            unit="batch",
            bar_format="{desc} ch{bar:20}nk {percentage:3.0f}% • {n_fmt}/{total_fmt} batches chunked [{elapsed}<{remaining}, {rate_fmt}] 🌱",
            ascii=" o",
        ):
            batch_texts = texts[i : min(i + batch_size, len(texts))]
            chunks.extend(self._process_batch(batch_texts))
        return chunks

    def __call__(  # type: ignore[override]
        self,
        text: Union[str, list[str]],
        batch_size: int = 1,
        show_progress_bar: bool = True,
    ) -> Union[list[Chunk], list[list[Chunk]]]:
        """Make the TokenChunker callable directly.

        Args:
            text: Input text or list of texts to be chunked
            batch_size: Number of texts to process in a single batch
            show_progress_bar: Whether to show a progress bar (for batch chunking)

        Returns:
            List of Chunk objects or list of lists of Chunk

        """
        if isinstance(text, str):
            return self.chunk(text)
        elif isinstance(text, list) and isinstance(text[0], str):
            return self.chunk_batch(text, batch_size, show_progress_bar)
        else:
            raise ValueError("Invalid input type. Expected a string or a list of strings.")

    def __repr__(self) -> str:
        """Return a string representation of the TokenChunker."""
        return (
            f"TokenChunker(tokenizer={self.tokenizer}, "
            f"chunk_size={self.chunk_size}, "
            f"chunk_overlap={self.chunk_overlap})"
        )



"""Slumber Chunking for Chonkie API."""

import os
from typing import Any, Optional, Union, cast

import httpx

from chonkie.cloud.file import FileManager
from chonkie.types import Chunk

from .base import CloudChunker


class SlumberChunker(CloudChunker):
    """Slumber Chunking for Chonkie API."""

    BASE_URL = "https://api.chonkie.ai"
    VERSION = "v1"

    def __init__(
        self,
        tokenizer: str = "gpt2",
        chunk_size: int = 1024,
        recipe: str = "default",
        lang: str = "en",
        candidate_size: int = 128,
        min_characters_per_chunk: int = 24,
        api_key: Optional[str] = None,
    ) -> None:
        """Initialize the SlumberChunker.

        Args:
            tokenizer (str): The tokenizer to use.
            chunk_size (int): The target size of the chunks.
            recipe (str): The recipe to use.
            lang (str): The language to use.
            candidate_size (int): The size of the candidate splits that the chunker will consider.
            min_characters_per_chunk (int): The minimum number of characters per chunk.
            api_key (Optional[str]): The Chonkie API key.

        """
        self.api_key = api_key or os.getenv("CHONKIE_API_KEY")
        if not self.api_key:
            raise ValueError(
                "No API key provided. Please set the CHONKIE_API_KEY environment variable"
                + " or pass an API key to the SlumberChunker constructor.",
            )

        if chunk_size <= 0:
            raise ValueError("Chunk size must be greater than 0.")
        if candidate_size <= 0:
            raise ValueError("Candidate size must be greater than 0.")
        if min_characters_per_chunk < 1:
            raise ValueError("Minimum characters per chunk must be greater than 0.")

        self.tokenizer = tokenizer
        self.chunk_size = chunk_size
        self.recipe = recipe
        self.lang = lang
        self.candidate_size = candidate_size
        self.min_characters_per_chunk = min_characters_per_chunk

        # Check if the API is up
        try:
            response = httpx.get(f"{self.BASE_URL}/")
            response.raise_for_status()  # Raises an HTTPStatusError for bad responses (4XX or 5XX)
        except httpx.HTTPError as error:
            raise ValueError(
                "Oh no! You caught Chonkie at a bad time. It seems to be down or unreachable."
                + " Please try again in a short while."
                + " If the issue persists, please contact support at support@chonkie.ai.",
            ) from error

        # Initialize the file manager to upload files if needed
        self.file_manager = FileManager(api_key=self.api_key)

    def chunk(
        self,
        text: Optional[Union[str, list[str]]] = None,
        file: str | os.PathLike | None = None,
    ) -> Union[list[Chunk], list[list[Chunk]]]:
        """Chunk the text or file into a list of chunks using the Slumber strategy via API.

        Args:
            text (Union[str, list[str]]): The text or list of texts to chunk.
            file (Optional[str]): The path to a file to chunk.

        Returns:
            list[Dict]: A list of dictionaries representing the chunks or texts.

        """
        payload: dict[str, Any]
        if text is not None:
            payload = {
                "text": text,
                "tokenizer_or_token_counter": self.tokenizer,
                "chunk_size": self.chunk_size,
                "recipe": self.recipe,
                "lang": self.lang,
                "candidate_size": self.candidate_size,
                "min_characters_per_chunk": self.min_characters_per_chunk,
            }
        elif file is not None:
            file_response = self.file_manager.upload(file)
            payload = {
                "file": {
                    "type": "document",
                    "content": file_response.name,
                },
                "tokenizer_or_token_counter": self.tokenizer,
                "chunk_size": self.chunk_size,
                "recipe": self.recipe,
                "lang": self.lang,
                "candidate_size": self.candidate_size,
                "min_characters_per_chunk": self.min_characters_per_chunk,
            }
        else:
            raise ValueError(
                "No text or file provided. Please provide either text or a file path.",
            )

        try:
            response = httpx.post(
                f"{self.BASE_URL}/{self.VERSION}/chunk/slumber",
                json=payload,
                headers={"Authorization": f"Bearer {self.api_key}"},
            )
            response.raise_for_status()  # Raises an HTTPStatusError for bad responses
        except httpx.HTTPError as e:
            # More specific error message including potential response text for debugging
            error_message = (
                "Oh no! The Chonkie API returned an error while trying to chunk with Slumber."
                + " Please try again in a short while."
            )
            resp = getattr(e, "response", None)
            if resp is not None:
                try:
                    error_detail = resp.json()
                    error_message += f" Details: {error_detail}"
                except ValueError:  # if response is not JSON
                    error_message += f" Status Code: {resp.status_code}. Response: {resp.text}"
            error_message += (
                " If the issue persists, please contact support at support@chonkie.ai."
            )
            raise ValueError(error_message) from e

        try:
            # Assuming the API always returns a list of dictionaries.
            if isinstance(text, list):
                batch_result: list[list[dict]] = cast(list[list[dict]], response.json())
                batch_chunks: list[list[Chunk]] = []
                for chunk_list in batch_result:
                    curr_chunks: list[Chunk] = []
                    for chunk in chunk_list:
                        curr_chunks.append(Chunk.from_dict(chunk))
                    batch_chunks.append(curr_chunks)
                return batch_chunks
            else:
                single_result: list[dict] = cast(list[dict], response.json())
                single_chunks: list[Chunk] = [Chunk.from_dict(chunk) for chunk in single_result]
                return single_chunks
        except ValueError as error:  # JSONDecodeError inherits from ValueError
            raise ValueError(
                "Oh no! The Chonkie API returned an invalid JSON response for Slumber chunking."
                + " Please try again in a short while."
                + f" Response text: {response.text}"
                + " If the issue persists, please contact support at support@chonkie.ai.",
            ) from error
        except Exception as error:  # Catch any other parsing/validation errors
            raise ValueError(
                "Oh no! Failed to parse the response from Chonkie API for Slumber chunking."
                + " Please try again in a short while."
                + f" Details: {str(error)}"
                + " If the issue persists, please contact support at support@chonkie.ai.",
            ) from error

    def __call__(
        self,
        text: Optional[Union[str, list[str]]] = None,
        file: str | os.PathLike | None = None,
    ) -> Union[list[Chunk], list[list[Chunk]]]:
        """Call the SlumberChunker."""
        return self.chunk(text=text, file=file)

    def __repr__(self) -> str:
        """Return a string representation of the SlumberChunker."""
        return (
            f"SlumberChunker(api_key={'********' if self.api_key else None}, "
            f"tokenizer='{self.tokenizer}', "
            f"chunk_size={self.chunk_size}, "
            f"recipe='{self.recipe}', "
            f"lang='{self.lang}', "
            f"candidate_size={self.candidate_size}, "
            f"min_characters_per_chunk={self.min_characters_per_chunk})"
        )



"""Chonkie: Recursive Chunker.

Splits text into smaller chunks recursively. Express chunking logic through RecursiveLevel objects.
"""

import os
from functools import lru_cache
from typing import Optional, Union

from chonkie.chunker.base import BaseChunker
from chonkie.logger import get_logger
from chonkie.pipeline import chunker
from chonkie.tokenizer import TokenizerProtocol
from chonkie.types import (
    Chunk,
    RecursiveLevel,
    RecursiveRules,
)

logger = get_logger(__name__)

import chonkie_core


@chunker("recursive")
class RecursiveChunker(BaseChunker):
    """Chunker that recursively splits text into smaller chunks, based on the provided RecursiveRules.

    Args:
        tokenizer: Tokenizer to use
        chunk_size (int): Maximum size of each chunk.
        rules: Recursive rules to use for chunking.
        min_characters_per_chunk (int): Minimum number of characters per chunk.

    """

    def __init__(
        self,
        tokenizer: Union[str, TokenizerProtocol] = "character",
        chunk_size: int = 2048,
        rules: RecursiveRules = RecursiveRules(),
        min_characters_per_chunk: int = 24,
    ) -> None:
        """Create a RecursiveChunker object.

        Args:
            tokenizer: Tokenizer to use
            chunk_size (int): Maximum size of each chunk.
            rules: Recursive rules to use for chunking.
            min_characters_per_chunk (int): Minimum number of characters per chunk.

        Raises:
            ValueError: If chunk_size <=0
            ValueError: If min_characters_per_chunk < 1
            ValueError: If rules is not a RecursiveRules object.

        """
        super().__init__(tokenizer=tokenizer)

        if chunk_size <= 0:
            raise ValueError("chunk_size must be greater than 0")
        if min_characters_per_chunk <= 0:
            raise ValueError("min_characters_per_chunk must be greater than 0")
        if not isinstance(rules, RecursiveRules):
            raise ValueError("`rules` must be a RecursiveRules object.")

        # Initialize the internal values
        self.chunk_size = chunk_size
        self.min_characters_per_chunk = min_characters_per_chunk
        self.rules = rules
        self.sep = "✄"
        self._CHARS_PER_TOKEN = 6.5

    @classmethod
    def from_recipe(
        cls,
        name: Optional[str] = "default",
        lang: Optional[str] = "en",
        path: str | os.PathLike | None = None,
        tokenizer: Union[str, TokenizerProtocol] = "character",
        chunk_size: int = 2048,
        min_characters_per_chunk: int = 24,
    ) -> "RecursiveChunker":
        """Create a RecursiveChunker object from a recipe.

        The recipes are registered in the [Chonkie Recipe Store](https://huggingface.co/datasets/chonkie-ai/recipes). If the recipe is not there, you can create your own recipe and share it with the community!

        Args:
            name (Optional[str]): The name of the recipe.
            lang (Optional[str]): The language that the recursive chunker should support.
            path (Optional[str]): The path to the recipe.
            tokenizer: The tokenizer to use.
            chunk_size (int): The chunk size.
            min_characters_per_chunk (int): The minimum number of characters per chunk.

        Returns:
            RecursiveChunker: The RecursiveChunker object.

        Raises:
            ValueError: If the recipe is not found.

        """
        logger.info("Loading RecursiveChunker recipe", recipe_name=name, lang=lang)
        # Create a recursive rules object
        rules = RecursiveRules.from_recipe(name, lang, path)
        logger.debug(f"Recipe loaded successfully with {len(rules.levels or [])} levels")
        return cls(
            tokenizer=tokenizer,
            rules=rules,
            chunk_size=chunk_size,
            min_characters_per_chunk=min_characters_per_chunk,
        )

    @lru_cache(maxsize=4096)
    def _estimate_token_count(self, text: str) -> int:
        # Always return the actual token count for accuracy
        # The estimate was only used as an optimization hint
        return self.tokenizer.count_tokens(text)

    def _split_text(self, text: str, recursive_level: RecursiveLevel) -> list[str]:
        """Split the text into chunks using the delimiters."""
        if recursive_level.delimiters:
            include_mode = recursive_level.include_delim or "prev"
            text_bytes = text.encode("utf-8")

            # Check if we have multi-byte delimiters
            delimiters = recursive_level.delimiters
            if isinstance(delimiters, str):
                delimiters = [delimiters]

            has_multibyte = any(len(d) > 1 for d in delimiters)

            if has_multibyte:
                # Use split_pattern_offsets for multi-byte patterns
                patterns = [d.encode("utf-8") for d in delimiters]
                offsets = chonkie_core.split_pattern_offsets(
                    text_bytes,
                    patterns=patterns,
                    include_delim=include_mode,
                    min_chars=self.min_characters_per_chunk,
                )
            else:
                # Use faster split_offsets for single-byte delimiters
                delim_bytes = "".join(delimiters).encode("utf-8")
                offsets = chonkie_core.split_offsets(
                    text_bytes,
                    delimiters=delim_bytes,
                    include_delim=include_mode,
                    min_chars=self.min_characters_per_chunk,
                )

            # Convert offsets to strings
            splits = [text_bytes[start:end].decode("utf-8") for start, end in offsets]
            return [s for s in splits if s]  # Filter empty strings
        elif recursive_level.whitespace:
            # Split on whitespace using split_offsets (preserves spaces for reconstruction)
            text_bytes = text.encode("utf-8")
            include_mode = recursive_level.include_delim or "prev"
            offsets = chonkie_core.split_offsets(
                text_bytes,
                delimiters=b" ",
                include_delim=include_mode,
                min_chars=self.min_characters_per_chunk,
            )
            splits = [text_bytes[start:end].decode("utf-8") for start, end in offsets]
            return [s for s in splits if s]
        else:
            # Encode, Split, and Decode
            encoded = self.tokenizer.encode(text)
            token_splits = [
                encoded[i : i + self.chunk_size] for i in range(0, len(encoded), self.chunk_size)
            ]
            splits = list(self.tokenizer.decode_batch(token_splits))
            return splits

    def _make_chunks(self, text: str, token_count: int, level: int, start_offset: int) -> Chunk:
        """Create a Chunk object with indices based on the current offset.

        This method calculates the start and end indices of the chunk using the provided start_offset and the length of the text,
        avoiding a slower full-text search for efficiency.

        Args:
            text (str): The text content of the chunk.
            token_count (int): The number of tokens in the chunk.
            level (int): The recursion level of the chunk.
            start_offset (int): The starting offset in the original text.

        Returns:
            Chunk: A chunk object with calculated start and end indices, text, and token count.

        """
        return Chunk(
            text=text,
            start_index=start_offset,
            end_index=start_offset + len(text),
            token_count=token_count,
        )

    def _merge_splits(
        self,
        splits: list[str],
        token_counts: list[int],
    ) -> tuple[list[str], list[int]]:
        """Merge short splits into larger chunks using chonkie-core."""
        if not splits or not token_counts:
            return [], []

        if len(splits) != len(token_counts):
            raise ValueError(
                f"Number of splits {len(splits)} does not match number of token counts {len(token_counts)}",
            )

        # If all splits are larger than the chunk size, return them
        if all(counts > self.chunk_size for counts in token_counts):
            return splits, token_counts

        # Use chonkie-core to merge (string joining done in Rust for performance)
        result = chonkie_core.merge_splits(splits, token_counts, self.chunk_size)
        return result.merged, result.token_counts

    def _recursive_chunk(self, text: str, level: int = 0, start_offset: int = 0) -> list[Chunk]:
        """Recursive helper for core chunking."""
        if not text:
            return []

        if level >= len(self.rules):
            return [self._make_chunks(text, self._estimate_token_count(text), level, start_offset)]

        curr_rule = self.rules[level]
        if curr_rule is None:
            return [self._make_chunks(text, self._estimate_token_count(text), level, start_offset)]

        splits = self._split_text(text, curr_rule)
        token_counts = [self._estimate_token_count(split) for split in splits]

        if curr_rule.delimiters is None and not curr_rule.whitespace:
            merged, combined_token_counts = splits, token_counts

        elif curr_rule.delimiters is None and curr_rule.whitespace:
            # With split_offsets using include_delim="prev", splits already contain trailing spaces
            # e.g., ["Hello ", "World ", "Test"] - so we just concatenate them
            merged, combined_token_counts = self._merge_splits(splits, token_counts)

        else:
            merged, combined_token_counts = self._merge_splits(splits, token_counts)

        # Chunk long merged splits
        chunks: list[Chunk] = []
        current_offset = start_offset
        for split, token_count in zip(merged, combined_token_counts):
            if token_count > self.chunk_size:
                recursive_result = self._recursive_chunk(split, level + 1, current_offset)
                chunks.extend(recursive_result)
            else:
                chunks.append(self._make_chunks(split, token_count, level, current_offset))
            # Update the offset by the length of the processed split.
            current_offset += len(split)
        return chunks

    def chunk(self, text: str) -> list[Chunk]:
        """Recursively chunk text.

        Args:
            text (str): Text to chunk.

        """
        logger.debug(f"Starting recursive chunking for text of length {len(text)}")
        chunks = self._recursive_chunk(text=text, level=0, start_offset=0)
        logger.info(f"Created {len(chunks)} chunks using recursive chunking")
        return chunks

    def __repr__(self) -> str:
        """Get a string representation of the recursive chunker."""
        return (
            f"RecursiveChunker(tokenizer={self.tokenizer},"
            f" rules={self.rules}, chunk_size={self.chunk_size}, "
            f"min_characters_per_chunk={self.min_characters_per_chunk})"
        )



"""Sentence Chunking for Chonkie API."""

import os
from typing import Any, Literal, Optional, Union, cast

import httpx

from chonkie.cloud.file import FileManager
from chonkie.types import Chunk

from .base import CloudChunker


class SentenceChunker(CloudChunker):
    """Sentence Chunking for Chonkie API."""

    BASE_URL = "https://api.chonkie.ai"
    VERSION = "v1"

    def __init__(
        self,
        tokenizer: str = "gpt2",
        chunk_size: int = 512,
        chunk_overlap: int = 0,
        min_sentences_per_chunk: int = 1,
        min_characters_per_sentence: int = 12,
        approximate: bool = True,
        delim: Union[str, list[str]] = [". ", "! ", "? ", "\n"],
        include_delim: Union[Literal["prev", "next"], None] = "prev",
        api_key: Optional[str] = None,
    ) -> None:
        """Initialize the SentenceChunker."""
        # If no API key is provided, use the environment variable
        self.api_key = api_key or os.getenv("CHONKIE_API_KEY")
        if not self.api_key:
            raise ValueError(
                "No API key provided. Please set the CHONKIE_API_KEY environment variable"
                + "or pass an API key to the SentenceChunker constructor.",
            )

        # Check if chunk_size and chunk_overlap are valid
        if chunk_size <= 0:
            raise ValueError("Chunk size must be greater than 0.")
        if chunk_overlap < 0:
            raise ValueError("Chunk overlap must be greater than or equal to 0.")
        if min_sentences_per_chunk < 1:
            raise ValueError("Minimum sentences per chunk must be greater than 0.")
        if min_characters_per_sentence < 1:
            raise ValueError("Minimum characters per sentence must be greater than 0.")
        if approximate not in [True, False]:
            raise ValueError("Approximate must be either True or False.")
        if include_delim not in ["prev", "next", None]:
            raise ValueError("Include delim must be either 'prev', 'next' or None.")

        # Assign all the attributes to the instance
        self.tokenizer = tokenizer
        self.chunk_size = chunk_size
        self.chunk_overlap = chunk_overlap
        self.min_sentences_per_chunk = min_sentences_per_chunk
        self.min_characters_per_sentence = min_characters_per_sentence
        self.approximate = approximate
        self.delim = delim
        self.include_delim = include_delim

        # Check if the API is up right now
        response = httpx.get(f"{self.BASE_URL}/")
        if response.status_code != 200:
            raise ValueError(
                "Oh no! You caught Chonkie at a bad time. It seems to be down right now."
                + "Please try again in a short while."
                + "If the issue persists, please contact support at support@chonkie.ai or raise an issue on GitHub.",
            )

        # Initialize the file manager to upload files if needed
        self.file_manager = FileManager(api_key=self.api_key)

    def chunk(
        self,
        text: Optional[Union[str, list[str]]] = None,
        file: str | os.PathLike | None = None,
    ) -> Union[list[Chunk], list[list[Chunk]]]:
        """Chunk the text or file via sentence boundaries."""
        # Define the payload for the request
        payload: dict[str, Any]
        if text is not None:
            payload = {
                "text": text,
                "tokenizer": self.tokenizer,
                "chunk_size": self.chunk_size,
                "chunk_overlap": self.chunk_overlap,
                "min_sentences_per_chunk": self.min_sentences_per_chunk,
                "min_characters_per_sentence": self.min_characters_per_sentence,
                "approximate": self.approximate,
                "delim": self.delim,
                "include_delim": self.include_delim,
            }
        elif file is not None:
            file_response = self.file_manager.upload(file)
            payload = {
                "file": {
                    "type": "document",
                    "content": file_response.name,
                },
                "tokenizer": self.tokenizer,
                "chunk_size": self.chunk_size,
                "chunk_overlap": self.chunk_overlap,
                "min_sentences_per_chunk": self.min_sentences_per_chunk,
                "min_characters_per_sentence": self.min_characters_per_sentence,
                "approximate": self.approximate,
                "delim": self.delim,
                "include_delim": self.include_delim,
            }
        else:
            raise ValueError(
                "No text or file provided. Please provide either text or a file path.",
            )

        # Make the request to the Chonkie API
        response = httpx.post(
            f"{self.BASE_URL}/{self.VERSION}/chunk/sentence",
            json=payload,
            headers={"Authorization": f"Bearer {self.api_key}"},
        )

        # Parse the response
        try:
            if isinstance(text, list):
                batch_result: list[list[dict]] = cast(list[list[dict]], response.json())
                batch_chunks: list[list[Chunk]] = []
                for chunk_list in batch_result:
                    curr_chunks: list[Chunk] = []
                    for chunk in chunk_list:
                        curr_chunks.append(Chunk.from_dict(chunk))
                    batch_chunks.append(curr_chunks)
                return batch_chunks
            else:
                single_result: list[dict] = cast(list[dict], response.json())
                single_chunks: list[Chunk] = [Chunk.from_dict(chunk) for chunk in single_result]
                return single_chunks
        except Exception as error:
            raise ValueError(
                "Oh no! The Chonkie API returned an invalid response."
                + "Please try again in a short while."
                + "If the issue persists, please contact support at support@chonkie.ai.",
            ) from error

    def __call__(
        self,
        text: Optional[Union[str, list[str]]] = None,
        file: str | os.PathLike | None = None,
    ) -> Union[list[Chunk], list[list[Chunk]]]:
        """Call the SentenceChunker."""
        return self.chunk(text=text, file=file)
